from os import getenv
from pathlib import Path

from pydantic import BaseModel
from xdg_base_dirs import xdg_config_home

from .env import JMP_CLIENT_CONFIG_HOME

CONFIG_API_VERSION = "jumpstarter.dev/v1alpha1"
CONFIG_PATH = Path(getenv(JMP_CLIENT_CONFIG_HOME, xdg_config_home() / "jumpstarter"))


class ObjectMeta(BaseModel):
    namespace: str | None
    name: str



from pathlib import Path

import pytest

from .common import ObjectMeta
from .exporter import ExporterConfigV1Alpha1, ExporterConfigV1Alpha1DriverInstance
from .tls import TLSConfigV1Alpha1

pytestmark = pytest.mark.anyio


def test_exporter_config(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setattr(ExporterConfigV1Alpha1, "BASE_PATH", tmp_path)

    path = tmp_path / "test.yaml"

    text = """apiVersion: jumpstarter.dev/v1alpha1
kind: ExporterConfig
metadata:
  namespace: default
  name: test
tls:
  ca: "cacertificatedata"
  insecure: true
endpoint: "jumpstarter.my-lab.com:1443"
token: "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz"

export:
  power:
    type: "jumpstarter_driver_power.driver.PduPower"
    config:
      host: "192.168.1.111"
      port: 1234
      auth:
          username: "admin"
          password: "secret"
  serial:
    type: "jumpstarter_driver_pyserial.driver.Pyserial"
    config:
      port: "/dev/ttyUSB0"
      baudrate: 115200
  nested:
    children:
      custom:
        type: "vendorpackage.CustomDriver"
        config:
          hello: "world"
"""
    path.write_text(
        text,
        encoding="utf-8",
    )

    config = ExporterConfigV1Alpha1.load("test")

    assert config == ExporterConfigV1Alpha1(
        alias="test",
        apiVersion="jumpstarter.dev/v1alpha1",
        kind="ExporterConfig",
        metadata=ObjectMeta(namespace="default", name="test"),
        endpoint="jumpstarter.my-lab.com:1443",
        token="dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        tls=TLSConfigV1Alpha1(ca="cacertificatedata", insecure=True),
        export={
            "power": ExporterConfigV1Alpha1DriverInstance(
                type="jumpstarter_driver_power.driver.PduPower",
                config={
                    "host": "192.168.1.111",
                    "port": 1234,
                    "auth": {
                        "username": "admin",
                        "password": "secret",
                    },
                },
            ),
            "serial": ExporterConfigV1Alpha1DriverInstance(
                type="jumpstarter_driver_pyserial.driver.Pyserial",
                config={
                    "port": "/dev/ttyUSB0",
                    "baudrate": 115200,
                },
            ),
            "nested": ExporterConfigV1Alpha1DriverInstance(
                children={
                    "custom": ExporterConfigV1Alpha1DriverInstance(
                        type="vendorpackage.CustomDriver",
                        children={},
                        config={
                            "hello": "world",
                        },
                    )
                },
            ),
        },
        config={},
        path=path,
    )

    path.unlink()

    ExporterConfigV1Alpha1.save(config)

    assert config == ExporterConfigV1Alpha1.load("test")



import logging
import socket
from ipaddress import ip_address


def get_ip_address(logger: logging.Logger | None = None) -> str:
    """Get the IP address of the host machine"""
    # Try to get the IP address using the hostname
    hostname = socket.gethostname()
    address = socket.gethostbyname(hostname)
    # If it returns nothing or a loopbadck address, do it the hard way
    if not address or ip_address(address).is_loopback:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
                s.connect(("192.175.48.1", 53))  # AS112
                return s.getsockname()[0]
        except Exception:
            if logger:
                logger.warning("Could not determine default IP address, falling back to 0.0.0.0")
            return "0.0.0.0"

    return address



import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml
from pydantic import ValidationError

from jumpstarter.common.exceptions import FileNotFoundError
from jumpstarter.config.client import ClientConfigV1Alpha1, ClientConfigV1Alpha1Drivers
from jumpstarter.config.common import ObjectMeta
from jumpstarter.config.env import JMP_DRIVERS_ALLOW, JMP_ENDPOINT, JMP_NAME, JMP_NAMESPACE, JMP_TOKEN


def test_client_ensure_exists_makes_dir(monkeypatch: pytest.MonkeyPatch):
    with tempfile.TemporaryDirectory() as d:
        monkeypatch.setattr(ClientConfigV1Alpha1, "CLIENT_CONFIGS_PATH", Path(d) / "clients")
        ClientConfigV1Alpha1.ensure_exists()
        assert os.path.exists(ClientConfigV1Alpha1.CLIENT_CONFIGS_PATH)


def test_client_config_try_from_env(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(JMP_NAMESPACE, "default")
    monkeypatch.setenv(JMP_NAME, "testclient")
    monkeypatch.setenv(JMP_TOKEN, "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz")
    monkeypatch.setenv(JMP_ENDPOINT, "jumpstarter.my-lab.com:1443")
    monkeypatch.setenv(JMP_DRIVERS_ALLOW, "jumpstarter.drivers.*,vendorpackage.*")

    config = ClientConfigV1Alpha1.try_from_env()
    assert config.alias == "default"
    assert config.metadata.namespace == "default"
    assert config.metadata.name == "testclient"
    assert config.token == "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz"
    assert config.endpoint == "jumpstarter.my-lab.com:1443"
    assert config.drivers.allow == ["jumpstarter.drivers.*", "vendorpackage.*"]
    assert config.drivers.unsafe is False


def test_client_config_try_from_env_not_set():
    config = ClientConfigV1Alpha1.try_from_env()
    assert config is None


def test_client_config_from_env(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(JMP_NAMESPACE, "default")
    monkeypatch.setenv(JMP_NAME, "testclient")
    monkeypatch.setenv(JMP_TOKEN, "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz")
    monkeypatch.setenv(JMP_ENDPOINT, "jumpstarter.my-lab.com:1443")
    monkeypatch.setenv(JMP_DRIVERS_ALLOW, "jumpstarter.drivers.*,vendorpackage.*")

    config = ClientConfigV1Alpha1.from_env()
    assert config.alias == "default"
    assert config.metadata.namespace == "default"
    assert config.metadata.name == "testclient"
    assert config.token == "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz"
    assert config.endpoint == "jumpstarter.my-lab.com:1443"
    assert config.drivers.allow == ["jumpstarter.drivers.*", "vendorpackage.*"]
    assert config.drivers.unsafe is False


def test_client_config_from_env_allow_unsafe(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(JMP_NAMESPACE, "default")
    monkeypatch.setenv(JMP_NAME, "testclient")
    monkeypatch.setenv(JMP_TOKEN, "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz")
    monkeypatch.setenv(JMP_ENDPOINT, "jumpstarter.my-lab.com:1443")
    monkeypatch.setenv(JMP_DRIVERS_ALLOW, "UNSAFE")

    config = ClientConfigV1Alpha1.from_env()
    assert config.alias == "default"
    assert config.metadata.namespace == "default"
    assert config.metadata.name == "testclient"
    assert config.token == "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz"
    assert config.endpoint == "jumpstarter.my-lab.com:1443"
    assert config.drivers.allow == []
    assert config.drivers.unsafe is True


@pytest.mark.parametrize("missing_field", [JMP_TOKEN, JMP_ENDPOINT])
def test_client_config_from_env_missing_field_raises(monkeypatch: pytest.MonkeyPatch, missing_field):
    monkeypatch.setenv(JMP_NAMESPACE, "default")
    monkeypatch.setenv(JMP_NAME, "testclient")
    monkeypatch.setenv(JMP_TOKEN, "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz")
    monkeypatch.setenv(JMP_ENDPOINT, "jumpstarter.my-lab.com:1443")
    monkeypatch.setenv(JMP_DRIVERS_ALLOW, "jumpstarter.drivers.*,vendorpackage.*")

    monkeypatch.delenv(missing_field)

    with pytest.raises(ValidationError):
        _ = ClientConfigV1Alpha1.from_env()


def test_client_config_from_file():
    CLIENT_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
metadata:
  namespace: default
  name: testclient
endpoint: jumpstarter.my-lab.com:1443
token: dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz
drivers:
  allow:
  - jumpstarter.drivers.*
  - vendorpackage.*
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(CLIENT_CONFIG)
        f.close()
        config = ClientConfigV1Alpha1.from_file(f.name)
        assert config.alias == f.name.split("/")[-1]
        assert config.metadata.namespace == "default"
        assert config.metadata.name == "testclient"
        assert config.endpoint == "jumpstarter.my-lab.com:1443"
        assert config.token == "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz"
        assert config.drivers.allow == ["jumpstarter.drivers.*", "vendorpackage.*"]
        assert config.drivers.unsafe is False
        os.unlink(f.name)


@pytest.mark.parametrize("invalid_field", ["apiVersion", "kind"])
def test_client_config_from_file_invalid_field_raises(invalid_field):
    CLIENT_CONFIG = {
        "apiVersion": "jumpstarter.dev/v1alpha1",
        "kind": "ClientConfig",
        "endpoint": "jumpstarter.my-lab.com:1443",
        "token": "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        "drivers": {"allow": ["jumpstarter.drivers.*", "vendorpackage.*"]},
    }

    CLIENT_CONFIG[invalid_field] = "foo"
    with tempfile.NamedTemporaryFile(mode="w") as f:
        yaml.safe_dump(CLIENT_CONFIG, f, sort_keys=False)
        with pytest.raises(ValueError):
            _ = ClientConfigV1Alpha1.from_file(f.name)


@pytest.mark.parametrize("missing_field", ["token", "endpoint", "drivers"])
def test_client_config_from_file_missing_field_raises(missing_field):
    CLIENT_CONFIG = {
        "apiVersion": "jumpstarter.dev/v1alpha1",
        "kind": "ClientConfig",
        "endpoint": "jumpstarter.my-lab.com:1443",
        "token": "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        "drivers": {"allow": ["jumpstarter.drivers.*", "vendorpackage.*"]},
    }

    del CLIENT_CONFIG[missing_field]
    with tempfile.NamedTemporaryFile(mode="w") as f:
        yaml.safe_dump(CLIENT_CONFIG, f, sort_keys=False)
        with pytest.raises(ValidationError):
            _ = ClientConfigV1Alpha1.from_file(f.name)


@pytest.mark.parametrize("invalid_field", ["allow"])
def test_client_config_from_file_invalid_drivers_field_raises(invalid_field):
    CLIENT_CONFIG = {
        "apiVersion": "jumpstarter.dev/v1alpha1",
        "kind": "ClientConfig",
        "endpoint": "jumpstarter.my-lab.com:1443",
        "token": "dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        "drivers": {"allow": ["jumpstarter.drivers.*", "vendorpackage.*"]},
    }

    CLIENT_CONFIG["drivers"][invalid_field] = "foo"
    with tempfile.NamedTemporaryFile(mode="w") as f:
        yaml.safe_dump(CLIENT_CONFIG, f, sort_keys=False)
        with pytest.raises(ValidationError):
            _ = ClientConfigV1Alpha1.from_file(f.name)


def test_client_config_load():
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write("")
        f.close()
        with patch.object(ClientConfigV1Alpha1, "_get_path", return_value=Path(f.name)) as get_path_mock:
            with patch.object(
                ClientConfigV1Alpha1,
                "from_file",
                return_value=ClientConfigV1Alpha1(
                    alias="another",
                    metadata=ObjectMeta(namespace="default", name="another"),
                    endpoint="abc",
                    token="123",
                    drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
                ),
            ) as from_file_mock:
                value = ClientConfigV1Alpha1.load("another")
                assert value.alias == "another"
                get_path_mock.assert_called_once_with("another")
                from_file_mock.assert_called_once_with(Path(f.name))
                os.unlink(f.name)


def test_client_config_load_not_found_raises():
    with pytest.raises(FileNotFoundError):
        _ = ClientConfigV1Alpha1.load("1235jklhbafsvd90u1234fsad")


def test_client_config_save(monkeypatch: pytest.MonkeyPatch):
    CLIENT_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
metadata:
  namespace: default
  name: testclient
endpoint: jumpstarter.my-lab.com:1443
tls:
  ca: ''
  insecure: false
token: dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz
grpcOptions: {}
drivers:
  allow:
  - jumpstarter.drivers.*
  - vendorpackage.*
  unsafe: false
"""
    config = ClientConfigV1Alpha1(
        alias="testclient",
        metadata=ObjectMeta(namespace="default", name="testclient"),
        endpoint="jumpstarter.my-lab.com:1443",
        token="dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        drivers=ClientConfigV1Alpha1Drivers(allow=["jumpstarter.drivers.*", "vendorpackage.*"], unsafe=False),
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        with patch.object(ClientConfigV1Alpha1, "_get_path", return_value=Path(f.name)) as _get_path_mock:
            with patch.object(ClientConfigV1Alpha1, "ensure_exists"):
                ClientConfigV1Alpha1.save(config)
                with open(f.name) as loaded:
                    value = loaded.read()
                    assert value == CLIENT_CONFIG
        _get_path_mock.assert_called_once_with("testclient")
        os.unlink(f.name)


def test_client_config_save_explicit_path():
    CLIENT_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
metadata:
  namespace: default
  name: testclient
endpoint: jumpstarter.my-lab.com:1443
tls:
  ca: ''
  insecure: false
token: dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz
grpcOptions: {}
drivers:
  allow:
  - jumpstarter.drivers.*
  - vendorpackage.*
  unsafe: false
"""
    config = ClientConfigV1Alpha1(
        alias="testclient",
        metadata=ObjectMeta(namespace="default", name="testclient"),
        endpoint="jumpstarter.my-lab.com:1443",
        token="dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        drivers=ClientConfigV1Alpha1Drivers(allow=["jumpstarter.drivers.*", "vendorpackage.*"], unsafe=False),
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        with patch.object(ClientConfigV1Alpha1, "ensure_exists"):
            ClientConfigV1Alpha1.save(config, f.name)
            with open(f.name) as loaded:
                value = loaded.read()
                assert value == CLIENT_CONFIG
        os.unlink(f.name)


def test_client_config_save_unsafe_drivers():
    CLIENT_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
metadata:
  namespace: default
  name: testclient
endpoint: jumpstarter.my-lab.com:1443
tls:
  ca: ''
  insecure: false
token: dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz
grpcOptions: {}
drivers:
  allow: []
  unsafe: true
"""
    config = ClientConfigV1Alpha1(
        alias="testclient",
        metadata=ObjectMeta(namespace="default", name="testclient"),
        endpoint="jumpstarter.my-lab.com:1443",
        token="dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz",
        drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=True),
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        with patch.object(ClientConfigV1Alpha1, "ensure_exists"):
            ClientConfigV1Alpha1.save(config, f.name)
            with open(f.name) as loaded:
                value = loaded.read()
                assert value == CLIENT_CONFIG
        os.unlink(f.name)


def test_client_config_exists():
    with patch.object(
        ClientConfigV1Alpha1, "_get_path", return_value=Path("/users/adsf/.config/jumpstarter/clients/abc.yaml")
    ) as _get_path_mock:
        assert ClientConfigV1Alpha1.exists("abc") is False
        _get_path_mock.assert_called_once_with("abc")


def test_client_config_list(monkeypatch: pytest.MonkeyPatch):
    CLIENT_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
metadata:
  namespace: default
  name: testclient
endpoint: jumpstarter.my-lab.com:1443
token: dGhpc2lzYXRva2VuLTEyMzQxMjM0MTIzNEyMzQtc2Rxd3Jxd2VycXdlcnF3ZXJxd2VyLTEyMzQxMjM0MTIz
drivers:
  allow:
  - jumpstarter.drivers.*
  - vendorpackage.*
"""
    with tempfile.TemporaryDirectory() as d:
        with open(Path(d) / "testclient.yaml", "w") as f:
            f.write(CLIENT_CONFIG)
            f.close()

        monkeypatch.setattr(ClientConfigV1Alpha1, "CLIENT_CONFIGS_PATH", Path(d))
        configs = ClientConfigV1Alpha1.list()
        assert len(configs) == 1
        assert configs[0].alias == "testclient"


def test_client_config_list_none(monkeypatch: pytest.MonkeyPatch):
    with tempfile.TemporaryDirectory() as d:
        monkeypatch.setattr(ClientConfigV1Alpha1, "CLIENT_CONFIGS_PATH", Path(d))
        configs = ClientConfigV1Alpha1.list()
        assert len(configs) == 0


def test_client_config_list_not_found_returns_empty(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(ClientConfigV1Alpha1, "CLIENT_CONFIGS_PATH", Path("/homeless-shelter"))
    configs = ClientConfigV1Alpha1.list()
    assert len(configs) == 0


def test_client_config_delete():
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        with patch.object(ClientConfigV1Alpha1, "_get_path", return_value=Path(f.name)) as _get_path_mock:
            f.write("")
            f.close()
            ClientConfigV1Alpha1.delete("testclient")
            _get_path_mock.assert_called_once_with("testclient")
            assert os.path.exists(f.name) is False


def test_client_config_delete_does_not_exist_raises():
    with patch.object(
        ClientConfigV1Alpha1, "_get_path", return_value=Path("/asdf/2134/cv/clients/xyz.yaml")
    ) as _get_path_mock:
        with pytest.raises(FileNotFoundError):
            ClientConfigV1Alpha1.delete("xyz")
        _get_path_mock.assert_called_once_with("xyz")



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta

import yaml
from google.protobuf import duration_pb2, field_mask_pb2, json_format
from grpc.aio import Channel
from jumpstarter_protocol import client_pb2, client_pb2_grpc, kubernetes_pb2
from pydantic import BaseModel, ConfigDict, Field, field_serializer

from jumpstarter.common.grpc import translate_grpc_exceptions


def parse_identifier(identifier: str, kind: str) -> (str, str):
    segments = identifier.split("/")
    if len(segments) != 4:
        raise ValueError("incorrect number of segments in identifier, expecting 4, got {}".format(len(segments)))
    if segments[0] != "namespaces":
        raise ValueError("incorrect first segment in identifier, expecting namespaces, got {}".format(segments[0]))
    if segments[2] != kind:
        raise ValueError("incorrect third segment in identifier, expecting {}, got {}".format(kind, segments[2]))
    return segments[1], segments[3]


def parse_client_identifier(identifier: str) -> (str, str):
    return parse_identifier(identifier, "clients")


def parse_exporter_identifier(identifier: str) -> (str, str):
    return parse_identifier(identifier, "exporters")


def parse_lease_identifier(identifier: str) -> (str, str):
    return parse_identifier(identifier, "leases")


class Exporter(BaseModel):
    namespace: str
    name: str
    labels: dict[str, str]

    @classmethod
    def from_protobuf(cls, data: client_pb2.Exporter) -> Exporter:
        namespace, name = parse_exporter_identifier(data.name)
        return cls(namespace=namespace, name=name, labels=data.labels)


class Lease(BaseModel):
    namespace: str
    name: str
    selector: str
    duration: timedelta
    client: str
    exporter: str
    conditions: list[kubernetes_pb2.Condition]
    effective_begin_time: datetime | None = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        ser_json_timedelta="float",
    )

    @field_serializer("conditions")
    def serialize_conditions(self, conditions: list[kubernetes_pb2.Condition], _info):
        return [json_format.MessageToDict(condition) for condition in conditions]

    @classmethod
    def from_protobuf(cls, data: client_pb2.Lease) -> Lease:
        namespace, name = parse_lease_identifier(data.name)

        _, client = parse_client_identifier(data.client)
        if data.exporter != "":
            _, exporter = parse_exporter_identifier(data.exporter)
        else:
            exporter = ""

        effective_begin_time = None
        if data.effective_begin_time:
            effective_begin_time = data.effective_begin_time.ToDatetime(
                tzinfo=datetime.now().astimezone().tzinfo,
            )

        return cls(
            namespace=namespace,
            name=name,
            selector=data.selector,
            duration=data.duration.ToTimedelta(),
            client=client,
            exporter=exporter,
            effective_begin_time=effective_begin_time,
            conditions=data.conditions,
        )

    def dump_json(self):
        return self.model_dump_json(indent=4, by_alias=True)

    def dump_yaml(self):
        return yaml.safe_dump(self.model_dump(mode="json", by_alias=True), indent=2)


class ExporterList(BaseModel):
    exporters: list[Exporter]
    next_page_token: str | None = Field(exclude=True)

    @classmethod
    def from_protobuf(cls, data: client_pb2.ListExportersResponse) -> ExporterList:
        return cls(
            exporters=list(map(Exporter.from_protobuf, data.exporters)),
            next_page_token=data.next_page_token,
        )

    def dump_json(self):
        return self.model_dump_json(indent=4, by_alias=True)

    def dump_yaml(self):
        return yaml.safe_dump(self.model_dump(mode="json", by_alias=True), indent=2)


class LeaseList(BaseModel):
    leases: list[Lease]
    next_page_token: str | None = Field(exclude=True)

    @classmethod
    def from_protobuf(cls, data: client_pb2.ListLeasesResponse) -> LeaseList:
        return cls(
            leases=list(map(Lease.from_protobuf, data.leases)),
            next_page_token=data.next_page_token,
        )

    def dump_json(self):
        return self.model_dump_json(indent=4, by_alias=True)

    def dump_yaml(self):
        return yaml.safe_dump(self.model_dump(mode="json", by_alias=True), indent=2)


@dataclass(kw_only=True, slots=True)
class ClientService:
    channel: Channel
    namespace: str
    stub: client_pb2_grpc.ClientServiceStub = field(init=False)

    def __post_init__(self):
        self.stub = client_pb2_grpc.ClientServiceStub(channel=self.channel)

    async def GetExporter(self, *, name: str):
        with translate_grpc_exceptions():
            exporter = await self.stub.GetExporter(
                client_pb2.GetExporterRequest(
                    name="namespaces/{}/exporters/{}".format(self.namespace, name),
                )
            )
        return Exporter.from_protobuf(exporter)

    async def ListExporters(
        self,
        *,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
    ):
        with translate_grpc_exceptions():
            exporters = await self.stub.ListExporters(
                client_pb2.ListExportersRequest(
                    parent="namespaces/{}".format(self.namespace),
                    page_size=page_size,
                    page_token=page_token,
                    filter=filter,
                )
            )
        return ExporterList.from_protobuf(exporters)

    async def GetLease(self, *, name: str):
        with translate_grpc_exceptions():
            lease = await self.stub.GetLease(
                client_pb2.GetLeaseRequest(
                    name="namespaces/{}/leases/{}".format(self.namespace, name),
                )
            )
        return Lease.from_protobuf(lease)

    async def ListLeases(
        self,
        *,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
    ):
        with translate_grpc_exceptions():
            leases = await self.stub.ListLeases(
                client_pb2.ListLeasesRequest(
                    parent="namespaces/{}".format(self.namespace),
                    page_size=page_size,
                    page_token=page_token,
                    filter=filter,
                )
            )
        return LeaseList.from_protobuf(leases)

    async def CreateLease(
        self,
        *,
        selector: str,
        duration: timedelta,
    ):
        duration_pb = duration_pb2.Duration()
        duration_pb.FromTimedelta(duration)

        with translate_grpc_exceptions():
            lease = await self.stub.CreateLease(
                client_pb2.CreateLeaseRequest(
                    parent="namespaces/{}".format(self.namespace),
                    lease=client_pb2.Lease(
                        duration=duration_pb,
                        selector=selector,
                    ),
                )
            )
        return Lease.from_protobuf(lease)

    async def UpdateLease(
        self,
        *,
        name: str,
        duration: timedelta,
    ):
        duration_pb = duration_pb2.Duration()
        duration_pb.FromTimedelta(duration)

        update_mask = field_mask_pb2.FieldMask()
        update_mask.FromJsonString("duration")

        with translate_grpc_exceptions():
            lease = await self.stub.UpdateLease(
                client_pb2.UpdateLeaseRequest(
                    lease=client_pb2.Lease(
                        name="namespaces/{}/leases/{}".format(self.namespace, name),
                        duration=duration_pb,
                    ),
                    update_mask=update_mask,
                )
            )
        return Lease.from_protobuf(lease)

    async def DeleteLease(self, *, name: str):
        with translate_grpc_exceptions():
            await self.stub.DeleteLease(
                client_pb2.DeleteLeaseRequest(
                    name="namespaces/{}/leases/{}".format(self.namespace, name),
                )
            )



"""
Base classes for drivers and driver clients
"""

import logging
from contextlib import asynccontextmanager
from dataclasses import dataclass, field

from anyio import create_task_group
from google.protobuf import empty_pb2
from grpc import StatusCode
from grpc.aio import AioRpcError, Channel
from jumpstarter_protocol import jumpstarter_pb2, jumpstarter_pb2_grpc, router_pb2_grpc

from jumpstarter.common import Metadata
from jumpstarter.common.exceptions import JumpstarterException
from jumpstarter.common.resources import ResourceMetadata
from jumpstarter.common.serde import decode_value, encode_value
from jumpstarter.common.streams import (
    DriverStreamRequest,
    ResourceStreamRequest,
    StreamRequestMetadata,
)
from jumpstarter.streams.common import forward_stream
from jumpstarter.streams.metadata import MetadataStream, MetadataStreamAttributes
from jumpstarter.streams.progress import ProgressStream
from jumpstarter.streams.router import RouterStream


class DriverError(JumpstarterException):
    """
    Raised when a driver call returns an error
    """


class DriverMethodNotImplemented(DriverError, NotImplementedError):
    """
    Raised when a driver method is not implemented
    """


class DriverInvalidArgument(DriverError, ValueError):
    """
    Raised when a driver method is called with invalid arguments
    """


@dataclass(kw_only=True)
class AsyncDriverClient(
    Metadata,
    jumpstarter_pb2_grpc.ExporterServiceStub,
    router_pb2_grpc.RouterServiceStub,
):
    """
    Async driver client base class

    Backing implementation of blocking driver client.
    """

    channel: Channel

    log_level: str = "INFO"
    logger: logging.Logger = field(init=False)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()
        jumpstarter_pb2_grpc.ExporterServiceStub.__init__(self, self.channel)
        router_pb2_grpc.RouterServiceStub.__init__(self, self.channel)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.log_level)

        # add default handler
        if not self.logger.handlers:
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter("%(name)s - %(levelname)s - %(message)s"))
            self.logger.addHandler(handler)

    async def call_async(self, method, *args):
        """Make DriverCall by method name and arguments"""

        request = jumpstarter_pb2.DriverCallRequest(
            uuid=str(self.uuid),
            method=method,
            args=[encode_value(arg) for arg in args],
        )

        try:
            response = await self.DriverCall(request)
        except AioRpcError as e:
            match e.code():
                case StatusCode.UNIMPLEMENTED:
                    raise DriverMethodNotImplemented(e.details()) from None
                case StatusCode.INVALID_ARGUMENT:
                    raise DriverInvalidArgument(e.details()) from None
                case StatusCode.UNKNOWN:
                    raise DriverError(e.details()) from None
                case _:
                    raise DriverError(e.details()) from e

        return decode_value(response.result)

    async def streamingcall_async(self, method, *args):
        """Make StreamingDriverCall by method name and arguments"""

        request = jumpstarter_pb2.StreamingDriverCallRequest(
            uuid=str(self.uuid),
            method=method,
            args=[encode_value(arg) for arg in args],
        )

        try:
            async for response in self.StreamingDriverCall(request):
                yield decode_value(response.result)
        except AioRpcError as e:
            match e.code():
                case StatusCode.UNIMPLEMENTED:
                    raise DriverMethodNotImplemented(e.details()) from None
                case StatusCode.INVALID_ARGUMENT:
                    raise DriverInvalidArgument(e.details()) from None
                case StatusCode.UNKNOWN:
                    raise DriverError(e.details()) from None
                case _:
                    raise DriverError(e.details()) from e

    @asynccontextmanager
    async def stream_async(self, method):
        context = self.Stream(
            metadata=StreamRequestMetadata.model_construct(request=DriverStreamRequest(uuid=self.uuid, method=method))
            .model_dump(mode="json", round_trip=True)
            .items(),
        )
        metadata = dict(list(await context.initial_metadata()))
        async with MetadataStream(stream=RouterStream(context=context), metadata=metadata) as stream:
            yield stream

    @asynccontextmanager
    async def resource_async(
        self,
        stream,
    ):
        context = self.Stream(
            metadata=StreamRequestMetadata.model_construct(request=ResourceStreamRequest(uuid=self.uuid))
            .model_dump(mode="json", round_trip=True)
            .items(),
        )
        metadata = dict(list(await context.initial_metadata()))
        async with MetadataStream(stream=RouterStream(context=context), metadata=metadata) as rstream:
            async with forward_stream(ProgressStream(stream=stream), rstream):
                yield ResourceMetadata(**rstream.extra(MetadataStreamAttributes.metadata)).resource.model_dump(
                    mode="json"
                )

    def __log(self, level: int, msg: str):
        self.logger.log(level, msg)

    @asynccontextmanager
    async def log_stream_async(self):
        async def log_stream():
            async for response in self.LogStream(empty_pb2.Empty()):
                self.__log(logging.getLevelName(response.severity), response.message)

        async with create_task_group() as tg:
            tg.start_soon(log_stream)
            try:
                yield
            finally:
                tg.cancel_scope.cancel()



import os
import sys
from contextlib import ExitStack, asynccontextmanager, contextmanager
from subprocess import Popen

from anyio.from_thread import BlockingPortal, start_blocking_portal

from jumpstarter.client import client_from_path
from jumpstarter.config.env import JMP_DRIVERS_ALLOW, JUMPSTARTER_HOST
from jumpstarter.driver import Driver
from jumpstarter.exporter import Session
from jumpstarter.utils.env import env

__all__ = ["env"]


@asynccontextmanager
async def serve_async(root_device: Driver, portal: BlockingPortal, stack: ExitStack):
    with Session(root_device=root_device) as session:
        async with session.serve_unix_async() as path:
            # SAFETY: the root_device instance is constructed locally thus considered trusted
            async with client_from_path(path, portal, stack, allow=[], unsafe=True) as client:
                try:
                    yield client
                finally:
                    if hasattr(client, "close"):
                        client.close()


@contextmanager
def serve(root_device: Driver):
    with start_blocking_portal() as portal:
        with ExitStack() as stack:
            with portal.wrap_async_context_manager(serve_async(root_device, portal, stack)) as client:
                try:
                    yield client
                finally:
                    if hasattr(client, "close"):
                        client.close()


ANSI_GRAY = "\\[\\e[90m\\]"
ANSI_YELLOW = "\\[\\e[93m\\]"
ANSI_WHITE = "\\[\\e[97m\\]"
ANSI_RESET = "\\[\\e[0m\\]"
PROMPT_CWD = "\\W"


def launch_shell(
    host: str,
    context: str,
    allow: [str],
    unsafe: bool,
    *,
    command: tuple[str, ...] | None = None,
) -> int:
    """Launch a shell with a custom prompt indicating the exporter type.

    Args:
        host: The jumpstarter host path
        context: The context of the shell ("local" or "remote")
        allow: List of allowed drivers
        unsafe: Whether to allow drivers outside of the allow list
    """

    env = os.environ | {
        JUMPSTARTER_HOST: host,
        JMP_DRIVERS_ALLOW: "UNSAFE" if unsafe else ",".join(allow),
        "PS1": f"{ANSI_GRAY}{PROMPT_CWD} {ANSI_YELLOW}⚡{ANSI_WHITE}{context} {ANSI_YELLOW}➤{ANSI_RESET} ",
    }

    if command:
        process = Popen(command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, env=env)
    else:
        cmd = [os.environ.get("SHELL", "bash")]
        if cmd[0].endswith("bash"):
            cmd.append("--norc")
            cmd.append("--noprofile")

        process = Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, env=env)

    return process.wait()



from dataclasses import dataclass

from anyio.abc import AnyByteStream
from anyio.from_thread import BlockingPortal


@dataclass(kw_only=True)
class BlockingStream:
    stream: AnyByteStream
    portal: BlockingPortal

    def send(self, data: bytes) -> None:
        return self.portal.call(self.stream.send, data)

    def receive(self) -> bytes:
        return self.portal.call(self.stream.receive)



import asyncio
import logging
from contextlib import asynccontextmanager, suppress

from anyio import (
    BrokenResourceError,
    ClosedResourceError,
    create_memory_object_stream,
    create_task_group,
)
from anyio.abc import AnyByteStream
from anyio.streams.stapled import StapledObjectStream

logger = logging.getLogger(__name__)


async def copy_stream(dst: AnyByteStream, src: AnyByteStream):
    with suppress(BrokenResourceError, ClosedResourceError, asyncio.exceptions.InvalidStateError):
        async for v in src:
            await dst.send(v)
        with suppress(
            AttributeError,
            # https://github.com/jumpstarter-dev/jumpstarter/issues/444
            # sending EOF to UDS on Darwin could result in
            # OSError: [Errno 57] Socket is not connected
            OSError,
        ):
            await dst.send_eof()


@asynccontextmanager
async def forward_stream(a, b):
    async with a, b:
        async with create_task_group() as tg:
            tg.start_soon(copy_stream, a, b)
            tg.start_soon(copy_stream, b, a)
            yield


def create_memory_stream():
    a_tx, a_rx = create_memory_object_stream[bytes](32)
    b_tx, b_rx = create_memory_object_stream[bytes](32)
    a = StapledObjectStream(a_tx, b_rx)
    b = StapledObjectStream(b_tx, a_rx)
    return a, b



from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Callable, Mapping

from anyio import TypedAttributeLookupError, TypedAttributeSet, typed_attribute
from anyio.abc import AnyByteStream, ObjectStream


class MetadataStreamAttributes(TypedAttributeSet):
    # https://grpc.io/docs/guides/metadata/
    metadata: dict[str, str] = typed_attribute()


@dataclass(frozen=True, kw_only=True, slots=True)
class MetadataStream(ObjectStream[bytes]):
    stream: AnyByteStream
    metadata: dict[str, str]

    async def send(self, item: bytes):
        await self.stream.send(item)

    async def receive(self) -> bytes:
        return await self.stream.receive()

    async def send_eof(self):
        await self.stream.send_eof()

    async def aclose(self):
        await self.stream.aclose()

    @property
    def extra_attributes(self) -> Mapping[Any, Callable[[], Any]]:
        metadata = {}
        with suppress(TypedAttributeLookupError):
            metadata = self.stream.extra(MetadataStreamAttributes.metadata)
        return self.stream.extra_attributes | {MetadataStreamAttributes.metadata: lambda: metadata | self.metadata}



from inspect import isasyncgenfunction, iscoroutinefunction, isfunction, isgeneratorfunction
from typing import Final

MARKER_MAGIC: Final[str] = "07c9b9cc"
MARKER_DRIVERCALL: Final[str] = "marker_drivercall"
MARKER_STREAMCALL: Final[str] = "marker_streamcall"
MARKER_STREAMING_DRIVERCALL: Final[str] = "marker_streamingdrivercall"


def export(func):
    """
    Decorator for exporting method as driver call
    """
    if isasyncgenfunction(func) or isgeneratorfunction(func):
        setattr(func, MARKER_STREAMING_DRIVERCALL, MARKER_MAGIC)
    elif iscoroutinefunction(func) or isfunction(func):
        setattr(func, MARKER_DRIVERCALL, MARKER_MAGIC)
    else:
        raise ValueError(f"unsupported exported function {func}")
    return func


def exportstream(func):
    """
    Decorator for exporting method as stream
    """
    setattr(func, MARKER_STREAMCALL, MARKER_MAGIC)
    return func



# Common environment variables for client/exporter config
JMP_CLIENT_CONFIG_HOME = "JMP_CLIENT_CONFIG_HOME"
JMP_CLIENT_CONFIG = "JMP_CLIENT_CONFIG"
JMP_NAMESPACE = "JMP_NAMESPACE"
JMP_NAME = "JMP_NAME"
JMP_ENDPOINT = "JMP_ENDPOINT"
JMP_TOKEN = "JMP_TOKEN"
JMP_DRIVERS_ALLOW = "JMP_DRIVERS_ALLOW"
JUMPSTARTER_HOST = "JUMPSTARTER_HOST"
JMP_LEASE = "JMP_LEASE"



import os
import tempfile
from unittest.mock import patch

import pytest

from jumpstarter.config.client import ClientConfigV1Alpha1, ClientConfigV1Alpha1Drivers
from jumpstarter.config.common import ObjectMeta
from jumpstarter.config.user import UserConfigV1Alpha1, UserConfigV1Alpha1Config


def test_user_config_exists(monkeypatch: pytest.MonkeyPatch):
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write("")
        f.close()
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        assert UserConfigV1Alpha1.exists() is True
        os.unlink(f.name)


def test_user_config_load(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
  current-client: testclient
"""
    with patch.object(
        ClientConfigV1Alpha1,
        "load",
        return_value=ClientConfigV1Alpha1(
            alias="testclient",
            metadata=ObjectMeta(namespace="default", name="testclient"),
            endpoint="abc",
            token="123",
            drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
        ),
    ) as mock_load:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(USER_CONFIG)
            f.close()
            monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
            config = UserConfigV1Alpha1.load()
            mock_load.assert_called_once_with("testclient")
            assert config.config.current_client.alias == "testclient"
            os.unlink(f.name)


def test_user_config_load_does_not_exist(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", "/nowhere/config.yaml`")
    with pytest.raises(FileNotFoundError):
        _ = UserConfigV1Alpha1.load()


def test_user_config_load_no_current_client(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config: {}
"""
    with patch.object(
        ClientConfigV1Alpha1,
        "load",
        return_value=ClientConfigV1Alpha1(
            name="testclient",
            metadata=ObjectMeta(namespace="default", name="testclient"),
            endpoint="abc",
            token="123",
            drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
        ),
    ) as mock_load:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(USER_CONFIG)
            f.close()
            monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
            config = UserConfigV1Alpha1.load()
            mock_load.assert_not_called()
            assert config.config.current_client is None
            os.unlink(f.name)


def test_user_config_load_current_client_empty(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
    current-client: null
"""
    with patch.object(
        ClientConfigV1Alpha1,
        "load",
        return_value=ClientConfigV1Alpha1(
            name="testclient",
            metadata=ObjectMeta(namespace="default", name="testclient"),
            endpoint="abc",
            token="123",
            drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
        ),
    ) as mock_load:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(USER_CONFIG)
            f.close()
            monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
            config = UserConfigV1Alpha1.load()
            mock_load.assert_not_called()
            assert config.config.current_client is None
            os.unlink(f.name)


def test_user_config_load_invalid_api_version_raises(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: abc
kind: UserConfig
config:
  current-client: null
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(USER_CONFIG)
        f.close()
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        with pytest.raises(ValueError):
            _ = UserConfigV1Alpha1.load()
        os.unlink(f.name)


def test_user_config_load_invalid_kind_raises(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: ClientConfig
config:
  current-client: null
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(USER_CONFIG)
        f.close()
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        with pytest.raises(ValueError):
            _ = UserConfigV1Alpha1.load()
        os.unlink(f.name)


def test_user_config_load_no_config_raises(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write(USER_CONFIG)
        f.close()
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        with pytest.raises(ValueError):
            _ = UserConfigV1Alpha1.load()
        os.unlink(f.name)


def test_user_config_load_or_create_config_exists():
    with patch.object(UserConfigV1Alpha1, "exists", return_value=True) as mock_exists:
        with patch.object(
            UserConfigV1Alpha1,
            "load",
            return_value=UserConfigV1Alpha1(config=UserConfigV1Alpha1Config(current_client=None)),
        ) as mock_load:
            _ = UserConfigV1Alpha1.load_or_create()
            mock_exists.assert_called_once()
            mock_load.assert_called_once()


def test_user_config_load_or_create_dir_exists():
    with patch.object(UserConfigV1Alpha1, "exists", return_value=False) as mock_exists:
        with patch.object(os.path, "exists", return_value=True):
            with patch.object(UserConfigV1Alpha1, "save") as mock_save:
                _ = UserConfigV1Alpha1.load_or_create()
                mock_exists.assert_called_once()
                mock_save.assert_called_once_with(
                    UserConfigV1Alpha1(config=UserConfigV1Alpha1Config(current_client=None))
                )


def test_user_config_load_or_create_dir_does_not_exist():
    with tempfile.TemporaryDirectory() as d:
        UserConfigV1Alpha1.BASE_CONFIG_PATH = f"{d}/jumpstarter"
        UserConfigV1Alpha1.USER_CONFIG_PATH = f"{d}/jumpstarter/config.yaml"
        with patch.object(UserConfigV1Alpha1, "save") as mock_save:
            _ = UserConfigV1Alpha1.load_or_create()
            mock_save.assert_called_once_with(UserConfigV1Alpha1(config=UserConfigV1Alpha1Config(current_client=None)))


def test_user_config_save(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
  current-client: testclient
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        config = UserConfigV1Alpha1(
            config=UserConfigV1Alpha1Config(
                current_client=ClientConfigV1Alpha1(
                    alias="testclient",
                    metadata=ObjectMeta(namespace="default", name="testclient"),
                    endpoint="abc",
                    token="123",
                    drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
                )
            )
        )
        UserConfigV1Alpha1.save(config)
        with open(f.name) as loaded:
            value = loaded.read()
            assert value == USER_CONFIG
        os.unlink(f.name)


def test_user_config_save_no_current_client(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
  current-client: null
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        config = UserConfigV1Alpha1(config=UserConfigV1Alpha1Config(current_client=None))
        UserConfigV1Alpha1.save(config)
        with open(f.name) as loaded:
            value = loaded.read()
            assert value == USER_CONFIG
        os.unlink(f.name)


def test_user_config_use_client(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
  current-client: testclient
"""
    with patch.object(
        ClientConfigV1Alpha1,
        "load",
        return_value=ClientConfigV1Alpha1(
            alias="testclient",
            metadata=ObjectMeta(namespace="default", name="testclient"),
            endpoint="abc",
            token="123",
            drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
        ),
    ) as mock_load:
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
            config = UserConfigV1Alpha1(
                config=UserConfigV1Alpha1Config(
                    current_client=ClientConfigV1Alpha1(
                        alias="another",
                        metadata=ObjectMeta(namespace="default", name="testclient"),
                        endpoint="abc",
                        token="123",
                        drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
                    )
                )
            )
            config.use_client("testclient")
            with open(f.name) as loaded:
                value = loaded.read()
                assert value == USER_CONFIG
                mock_load.assert_called_once_with("testclient")
            assert config.config.current_client.alias == "testclient"
            os.unlink(f.name)


def test_user_config_use_client_none(monkeypatch: pytest.MonkeyPatch):
    USER_CONFIG = """apiVersion: jumpstarter.dev/v1alpha1
kind: UserConfig
config:
  current-client: null
"""
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        monkeypatch.setattr(UserConfigV1Alpha1, "USER_CONFIG_PATH", f.name)
        config = UserConfigV1Alpha1(
            config=UserConfigV1Alpha1Config(
                current_client=ClientConfigV1Alpha1(
                    alias="another",
                    metadata=ObjectMeta(namespace="default", name="testclient"),
                    endpoint="abc",
                    token="123",
                    drivers=ClientConfigV1Alpha1Drivers(allow=[], unsafe=False),
                )
            )
        )
        config.use_client(None)
        with open(f.name) as loaded:
            value = loaded.read()
            assert value == USER_CONFIG
        assert config.config.current_client is None
        os.unlink(f.name)



import os
from contextlib import asynccontextmanager, contextmanager
from datetime import timedelta
from pathlib import Path
from typing import ClassVar, Literal, Optional, Self

import grpc
import yaml
from anyio.from_thread import BlockingPortal, start_blocking_portal
from pydantic import BaseModel, ConfigDict, Field, ValidationError

from .common import CONFIG_PATH, ObjectMeta
from .env import JMP_DRIVERS_ALLOW, JMP_ENDPOINT, JMP_LEASE, JMP_NAME, JMP_NAMESPACE, JMP_TOKEN
from .grpc import call_credentials
from .tls import TLSConfigV1Alpha1
from jumpstarter.client.grpc import ClientService
from jumpstarter.common.exceptions import FileNotFoundError
from jumpstarter.common.grpc import aio_secure_channel, ssl_channel_credentials


def _allow_from_env():
    allow = os.environ.get(JMP_DRIVERS_ALLOW)
    match allow:
        case None:
            return [], False
        case "UNSAFE":
            return [], True
        case _:
            return allow.split(","), False


class ClientConfigV1Alpha1Drivers(BaseModel):
    allow: list[str] = Field(default_factory=[])
    unsafe: bool = Field(default=False)


class ClientConfigV1Alpha1(BaseModel):
    CLIENT_CONFIGS_PATH: ClassVar[Path] = CONFIG_PATH / "clients"

    alias: str = Field(default="default")
    path: Path | None = Field(default=None)

    apiVersion: Literal["jumpstarter.dev/v1alpha1"] = Field(default="jumpstarter.dev/v1alpha1")
    kind: Literal["ClientConfig"] = Field(default="ClientConfig")

    metadata: ObjectMeta

    endpoint: str
    tls: TLSConfigV1Alpha1 = Field(default_factory=TLSConfigV1Alpha1)
    token: str
    grpcOptions: dict[str, str | int] | None = Field(default_factory=dict)

    drivers: ClientConfigV1Alpha1Drivers

    async def channel(self):
        credentials = grpc.composite_channel_credentials(
            await ssl_channel_credentials(self.endpoint, self.tls),
            call_credentials("Client", self.metadata, self.token),
        )

        return aio_secure_channel(self.endpoint, credentials, self.grpcOptions)

    @contextmanager
    def lease(
        self,
        selector: str | None = None,
        lease_name: str | None = None,
        duration: timedelta = timedelta(minutes=30),
    ):
        with start_blocking_portal() as portal:
            with portal.wrap_async_context_manager(self.lease_async(selector, lease_name, duration, portal)) as lease:
                yield lease

    def get_exporter(self, name: str):
        with start_blocking_portal() as portal:
            return portal.call(self.get_exporter_async, name)

    def list_exporters(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
    ):
        with start_blocking_portal() as portal:
            return portal.call(self.list_exporters_async, page_size, page_token, filter)

    def list_leases(self, filter: str):
        with start_blocking_portal() as portal:
            return portal.call(self.list_leases_async, filter)

    def create_lease(
        self,
        selector: str,
        duration: timedelta,
    ):
        with start_blocking_portal() as portal:
            return portal.call(self.create_lease_async, selector, duration)

    def delete_lease(
        self,
        name: str,
    ):
        with start_blocking_portal() as portal:
            return portal.call(self.delete_lease_async, name)

    def update_lease(self, name, duration: timedelta):
        with start_blocking_portal() as portal:
            return portal.call(self.update_lease_async, name, duration)

    async def get_exporter_async(self, name: str):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        return await svc.GetExporter(name=name)

    async def list_exporters_async(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
    ):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        return await svc.ListExporters(page_size=page_size, page_token=page_token, filter=filter)

    async def create_lease_async(
        self,
        selector: str,
        duration: timedelta,
    ):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        return await svc.CreateLease(
            selector=selector,
            duration=duration,
        )

    async def delete_lease_async(self, name: str):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        await svc.DeleteLease(
            name=name,
        )

    async def list_leases_async(self, filter: str):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        return await svc.ListLeases(filter=filter)

    async def update_lease_async(self, name, duration: timedelta):
        svc = ClientService(channel=await self.channel(), namespace=self.metadata.namespace)
        return await svc.UpdateLease(name=name, duration=duration)

    @asynccontextmanager
    async def lease_async(
        self,
        selector: str,
        lease_name: str | None,
        duration: timedelta,
        portal: BlockingPortal,
    ):
        from jumpstarter.client import Lease

        # if no lease_name provided, check if it is set in the environment
        lease_name = lease_name or os.environ.get(JMP_LEASE, "")
        # when no lease name is provided, release the lease on exit
        release_lease = lease_name == ""

        async with Lease(
            channel=await self.channel(),
            namespace=self.metadata.namespace,
            name=lease_name,
            selector=selector,
            duration=duration,
            portal=portal,
            allow=self.drivers.allow,
            unsafe=self.drivers.unsafe,
            release=release_lease,
            tls_config=self.tls,
            grpc_options=self.grpcOptions,
        ) as lease:
            yield lease

    @classmethod
    def from_file(cls, path: os.PathLike):
        with open(path) as f:
            v = cls.model_validate(yaml.safe_load(f))
            v.alias = os.path.basename(path).split(".")[0]
            v.path = Path(path)
            return v

    @classmethod
    def ensure_exists(cls):
        """Check if the clients config dir exists, otherwise create it."""
        os.makedirs(cls.CLIENT_CONFIGS_PATH, exist_ok=True)

    @classmethod
    def try_from_env(cls):
        try:
            return cls.from_env()
        except ValidationError:
            return None

    @classmethod
    def from_env(cls):
        allow, unsafe = _allow_from_env()
        return cls(
            metadata=ObjectMeta(
                namespace=os.environ.get(JMP_NAMESPACE),
                name=os.environ.get(JMP_NAME),
            ),
            endpoint=os.environ.get(JMP_ENDPOINT),
            token=os.environ.get(JMP_TOKEN),
            drivers=ClientConfigV1Alpha1Drivers(
                allow=allow,
                unsafe=unsafe,
            ),
        )

    @classmethod
    def _get_path(cls, alias: str) -> Path:
        """Get the regular path of a client config given an alias."""
        return (cls.CLIENT_CONFIGS_PATH / alias).with_suffix(".yaml")

    @classmethod
    def load(cls, alias: str) -> Self:
        """Load a client config by alias."""
        path = cls._get_path(alias)
        if path.exists() is False:
            raise FileNotFoundError(f"Client config '{path}' does not exist.")
        return cls.from_file(path)

    @classmethod
    def save(cls, config: Self, path: Optional[os.PathLike] = None) -> Path:
        """Saves a client config as YAML."""
        # Ensure the clients dir exists
        if path is None:
            cls.ensure_exists()
            # Set the config path before saving
            config.path = cls._get_path(config.alias)
        else:
            config.path = Path(path)
        with config.path.open(mode="w") as f:
            yaml.safe_dump(config.model_dump(mode="json", exclude={"path", "alias"}), f, sort_keys=False)
        return config.path

    @classmethod
    def dump_yaml(cls, config: Self) -> str:
        return yaml.safe_dump(config.model_dump(mode="json", exclude={"path", "alias"}), sort_keys=False)

    @classmethod
    def exists(cls, alias: str) -> bool:
        """Check if a client config exists by alias."""
        return cls._get_path(alias).exists()

    @classmethod
    def list(cls) -> list[Self]:
        """List the available client configs."""
        if cls.CLIENT_CONFIGS_PATH.exists() is False:
            # Return an empty list if the dir does not exist
            return []

        results = os.listdir(cls.CLIENT_CONFIGS_PATH)
        # Only accept YAML files in the list
        files = filter(lambda x: x.endswith(".yaml"), results)

        def make_config(file: str):
            path = cls.CLIENT_CONFIGS_PATH / file
            return cls.from_file(path)

        return list(map(make_config, files))

    @classmethod
    def delete(cls, alias: str) -> Path:
        """Delete a client config by alias."""
        path = cls._get_path(alias)
        if path.exists() is False:
            raise FileNotFoundError(f"Client config '{path}' does not exist.")
        path.unlink()
        return path


class ClientConfigListV1Alpha1(BaseModel):
    api_version: Literal["jumpstarter.dev/v1alpha1"] = Field(alias="apiVersion", default="jumpstarter.dev/v1alpha1")
    current_config: Optional[str] = Field(alias="currentConfig")
    items: list[ClientConfigV1Alpha1]
    kind: Literal["ClientConfigList"] = Field(default="ClientConfigList")

    def dump_json(self):
        return self.model_dump_json(indent=4, by_alias=True)

    def dump_yaml(self):
        return yaml.safe_dump(self.model_dump(mode="json", by_alias=True), indent=2)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)



"""
Base classes for drivers and driver clients
"""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from dataclasses import field
from inspect import isasyncgenfunction, iscoroutinefunction
from itertools import chain
from typing import Any
from uuid import UUID, uuid4

import aiohttp
from anyio import to_thread
from grpc import StatusCode
from jumpstarter_protocol import jumpstarter_pb2, jumpstarter_pb2_grpc, router_pb2_grpc
from pydantic import TypeAdapter
from pydantic.dataclasses import dataclass

from .decorators import (
    MARKER_DRIVERCALL,
    MARKER_MAGIC,
    MARKER_STREAMCALL,
    MARKER_STREAMING_DRIVERCALL,
)
from jumpstarter.common import Metadata
from jumpstarter.common.resources import ClientStreamResource, PresignedRequestResource, Resource, ResourceMetadata
from jumpstarter.common.serde import decode_value, encode_value
from jumpstarter.common.streams import (
    DriverStreamRequest,
    ResourceStreamRequest,
)
from jumpstarter.streams.aiohttp import AiohttpStreamReaderStream
from jumpstarter.streams.common import create_memory_stream
from jumpstarter.streams.metadata import MetadataStream
from jumpstarter.streams.progress import ProgressStream


@dataclass(kw_only=True)
class Driver(
    Metadata,
    jumpstarter_pb2_grpc.ExporterServiceServicer,
    router_pb2_grpc.RouterServiceServicer,
    metaclass=ABCMeta,
):
    """Base class for drivers

    Drivers should at the minimum implement the `client` method.

    Regular or streaming driver calls can be marked with the `export` decorator.
    Raw stream constructors can be marked with the `exportstream` decorator.
    """

    children: dict[str, Driver] = field(default_factory=dict)

    resources: dict[UUID, Any] = field(default_factory=dict, init=False)
    """Dict of client side resources"""

    log_level: str = "INFO"
    logger: logging.Logger = field(init=False)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.log_level)

    def close(self):
        for child in self.children.values():
            child.close()

    def reset(self):
        for child in self.children.values():
            child.reset()

    @classmethod
    @abstractmethod
    def client(cls) -> str:
        """
        Return full import path of the corresponding driver client class
        """

    def extra_labels(self) -> dict[str, str]:
        return {}

    async def DriverCall(self, request, context):
        """
        :meta private:
        """
        try:
            method = await self.__lookup_drivercall(request.method, context, MARKER_DRIVERCALL)

            args = [decode_value(arg) for arg in request.args]

            if iscoroutinefunction(method):
                result = await method(*args)
            else:
                result = await to_thread.run_sync(method, *args)

            return jumpstarter_pb2.DriverCallResponse(
                uuid=str(uuid4()),
                result=encode_value(result),
            )
        except NotImplementedError as e:
            await context.abort(StatusCode.UNIMPLEMENTED, str(e))
        except ValueError as e:
            await context.abort(StatusCode.INVALID_ARGUMENT, str(e))
        except TimeoutError as e:
            await context.abort(StatusCode.DEADLINE_EXCEEDED, str(e))
        except Exception as e:
            await context.abort(StatusCode.UNKNOWN, str(e))

    async def StreamingDriverCall(self, request, context):
        """
        :meta private:
        """
        try:
            method = await self.__lookup_drivercall(request.method, context, MARKER_STREAMING_DRIVERCALL)

            args = [decode_value(arg) for arg in request.args]

            if isasyncgenfunction(method):
                async for result in method(*args):
                    yield jumpstarter_pb2.StreamingDriverCallResponse(
                        uuid=str(uuid4()),
                        result=encode_value(result),
                    )
            else:
                for result in await to_thread.run_sync(method, *args):
                    yield jumpstarter_pb2.StreamingDriverCallResponse(
                        uuid=str(uuid4()),
                        result=encode_value(result),
                    )
        except NotImplementedError as e:
            await context.abort(StatusCode.UNIMPLEMENTED, str(e))
        except ValueError as e:
            await context.abort(StatusCode.INVALID_ARGUMENT, str(e))
        except TimeoutError as e:
            await context.abort(StatusCode.DEADLINE_EXCEEDED, str(e))
        except Exception as e:
            await context.abort(StatusCode.UNKNOWN, str(e))

    @asynccontextmanager
    async def Stream(self, request, context):
        """
        :meta private:
        """
        match request:
            case DriverStreamRequest(method=driver_method):
                method = await self.__lookup_drivercall(driver_method, context, MARKER_STREAMCALL)

                async with method() as stream:
                    yield stream

            case ResourceStreamRequest():
                remote, resource = create_memory_stream()

                resource_uuid = uuid4()

                self.resources[resource_uuid] = resource

                async with MetadataStream(
                    stream=remote,
                    metadata=ResourceMetadata.model_construct(
                        resource=ClientStreamResource(uuid=resource_uuid)
                    ).model_dump(mode="json", round_trip=True),
                ) as stream:
                    yield stream

    def report(self, *, root=None, parent=None, name=None):
        """
        Create DriverInstanceReport

        :meta private:
        """

        if root is None:
            root = self

        return jumpstarter_pb2.DriverInstanceReport(
            uuid=str(self.uuid),
            parent_uuid=str(parent.uuid) if parent else None,
            labels=self.labels
            | self.extra_labels()
            | ({"jumpstarter.dev/client": self.client()})
            | ({"jumpstarter.dev/name": name} if name else {}),
        )

    def enumerate(self, *, root=None, parent=None, name=None):
        """
        Get list of self and child devices

        :meta private:
        """
        if root is None:
            root = self

        return [(self.uuid, parent, name, self)] + list(
            chain(*[child.enumerate(root=root, parent=self, name=cname) for (cname, child) in self.children.items()])
        )

    @asynccontextmanager
    async def resource(self, handle: str, timeout: int = 300):
        handle = TypeAdapter(Resource).validate_python(handle)
        match handle:
            case ClientStreamResource(uuid=uuid):
                async with self.resources[uuid] as stream:
                    try:
                        yield stream
                    finally:
                        del self.resources[uuid]
            case PresignedRequestResource(headers=headers, url=url, method=method):
                client_timeout = aiohttp.ClientTimeout(total=timeout)
                match method:
                    case "GET":
                        async with aiohttp.request(
                            method, url, headers=headers, raise_for_status=True, timeout=client_timeout
                        ) as resp:
                            async with AiohttpStreamReaderStream(reader=resp.content) as stream:
                                yield ProgressStream(stream=stream, logging=True)
                    case "PUT":
                        remote, stream = create_memory_stream()
                        async with aiohttp.request(
                            method, url, headers=headers, raise_for_status=True, data=remote, timeout=client_timeout
                        ) as resp:
                            async with stream:
                                yield ProgressStream(stream=stream, logging=True)
                    case _:
                        # INVARIANT: method is always one of GET or PUT, see PresignedRequestResource
                        raise ValueError("unreachable")

    async def __lookup_drivercall(self, name, context, marker):
        """Lookup drivercall by method name

        Methods are checked against magic markers
        to avoid accidentally calling non-exported
        methods
        """
        method = getattr(self, name, None)

        if method is None:
            await context.abort(StatusCode.NOT_FOUND, f"method {name} not found on driver")

        if getattr(method, marker, None) != MARKER_MAGIC:
            await context.abort(StatusCode.NOT_FOUND, f"method {name} missing marker {marker}")

        return method



"""
Base classes for drivers and driver clients
"""

from __future__ import annotations

from contextlib import ExitStack, contextmanager
from dataclasses import field

from anyio.from_thread import BlockingPortal
from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from .core import AsyncDriverClient
from jumpstarter.streams.blocking import BlockingStream


@dataclass(kw_only=True, config=ConfigDict(arbitrary_types_allowed=True))
class DriverClient(AsyncDriverClient):
    """Base class for driver clients

    Client methods can be implemented as regular functions,
    and call the `call` or `streamingcall` helpers internally
    to invoke exported methods on the driver.

    Additional client functionalities such as raw stream
    connections or sharing client-side resources can be added
    by inheriting mixin classes under `jumpstarter.drivers.mixins`
    """

    children: dict[str, DriverClient] = field(default_factory=dict)

    portal: BlockingPortal
    stack: ExitStack

    def call(self, method, *args):
        """
        Invoke driver call

        :param str method: method name of driver call
        :param list[Any] args: arguments for driver call

        :return: driver call result
        :rtype: Any
        """
        return self.portal.call(self.call_async, method, *args)

    def streamingcall(self, method, *args):
        """
        Invoke streaming driver call

        :param str method: method name of streaming driver call
        :param list[Any] args: arguments for streaming driver call

        :return: streaming driver call result
        :rtype: Generator[Any, None, None]
        """
        generator = self.portal.call(self.streamingcall_async, method, *args)
        while True:
            try:
                yield self.portal.call(generator.__anext__)
            except StopAsyncIteration:
                break

    @contextmanager
    def stream(self, method="connect"):
        """
        Open a blocking stream session with a context manager.

        :param str method: method name of streaming driver call

        :return: blocking stream session object context manager.
        """

        with self.portal.wrap_async_context_manager(self.stream_async(method)) as stream:
            yield BlockingStream(stream=stream, portal=self.portal)

    @contextmanager
    def log_stream(self):
        with self.portal.wrap_async_context_manager(self.log_stream_async()):
            yield

    def open_stream(self) -> BlockingStream:
        """
        Open a blocking stream session without a context manager.

        :return: blocking stream session object.
        :rtype: BlockingStream
        """
        return self.stack.enter_context(self.stream())

    def close(self):
        """
        Close the open stream session without a context manager.
        """
        self.stack.close()

    def __del__(self):
        self.close()



import asyncio
import contextlib
import logging
from dataclasses import dataclass, field

import grpc
from anyio import (
    BrokenResourceError,
    EndOfStream,
)
from anyio.abc import ObjectStream
from jumpstarter_protocol import router_pb2

logger = logging.getLogger(__name__)


@dataclass(kw_only=True, slots=True)
class RouterStream(ObjectStream[bytes]):
    context: grpc.aio.StreamStreamCall | grpc._cython.cygrpc._ServicerContext
    cls: type = field(init=False)

    def __post_init__(self):
        match self.context:
            case grpc.aio.StreamStreamCall():
                self.cls = router_pb2.StreamRequest
            case grpc._cython.cygrpc._ServicerContext():
                self.cls = router_pb2.StreamResponse
            case _:
                raise ValueError(f"RouterStream: invalid context type: {type(self.context)}")

    async def send(self, payload: bytes) -> None:
        try:
            await self.context.write(self.cls(payload=payload))
        except grpc.aio.AioRpcError as e:
            raise BrokenResourceError from e

    async def receive(self) -> bytes:
        try:
            frame = await self.context.read()
        except grpc.aio.AioRpcError as e:
            raise BrokenResourceError from e

        # Reference: https://grpc.github.io/grpc/python/grpc_asyncio.html#grpc.aio.StreamStreamCall.read
        if frame == grpc.aio.EOF:
            raise EndOfStream

        match frame.frame_type:
            case router_pb2.FRAME_TYPE_DATA:
                return frame.payload
            case router_pb2.FRAME_TYPE_GOAWAY:
                raise EndOfStream
            case router_pb2.FRAME_TYPE_PING:
                pass
            case _:
                logger.debug(f"RouterStream: unrecognized frame ignored: {frame}")

        return b""

    async def send_eof(self):
        with contextlib.suppress(grpc.aio.AioRpcError, asyncio.exceptions.InvalidStateError):
            await self.context.write(self.cls(frame_type=router_pb2.FRAME_TYPE_GOAWAY))
            if isinstance(self.context, grpc.aio.StreamStreamCall):
                await self.context.done_writing()

    async def aclose(self):
        with contextlib.suppress(grpc.aio.AioRpcError, asyncio.exceptions.InvalidStateError):
            await self.send_eof()
            if isinstance(self.context, grpc._cython.cygrpc._ServicerContext):
                await self.context.abort(grpc.StatusCode.ABORTED, "RouterStream: aclose")



import asyncio
import base64
import os
import socket
import ssl
from contextlib import contextmanager
from typing import Any, Sequence, Tuple
from urllib.parse import urlparse

import grpc
from anyio import fail_after

from jumpstarter.common.exceptions import ConfigurationError, ConnectionError


async def ssl_channel_credentials(target: str, tls_config, timeout=5):
    configure_grpc_env()
    if tls_config.insecure or os.getenv("JUMPSTARTER_GRPC_INSECURE") == "1":
        try:
            parsed = urlparse(f"//{target}")
            port = parsed.port if parsed.port else 443
        except ValueError as e:
            raise ConfigurationError(f"Failed parsing {target}") from e

        try:
            with fail_after(timeout):
                ssl_context = ssl.create_default_context()
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE
                _, writer = await asyncio.open_connection(parsed.hostname, port, ssl=ssl_context)
                root_certificates = ""
                for cert in writer.get_extra_info("ssl_object")._sslobj.get_unverified_chain():
                    root_certificates += cert.public_bytes()
            return grpc.ssl_channel_credentials(root_certificates=root_certificates.encode())
        except socket.gaierror as e:
            raise ConnectionError(f"Failed resolving {parsed.hostname}") from e
        except ConnectionRefusedError as e:
            raise ConnectionError(f"Failed connecting to {parsed.hostname}:{port}") from e
        except TimeoutError as e:
            raise ConnectionError(f"Timeout connecting to {parsed.hostname}:{port}") from e

    elif tls_config.ca != "":
        ca_certificate = base64.b64decode(tls_config.ca)
        return grpc.ssl_channel_credentials(ca_certificate)
    else:
        return grpc.ssl_channel_credentials()


def aio_secure_channel(target: str, credentials: grpc.ChannelCredentials, grpc_options: dict[str, Any] | None):
    return grpc.aio.secure_channel(
        target,
        credentials,
        options=_override_default_grpc_options(grpc_options),
    )


def _override_default_grpc_options(grpc_options: dict[str, str | int] | None) -> Sequence[Tuple[str, Any]]:
    defaults = (
        ("grpc.lb_policy_name", "round_robin"),
        # we keep a low keepalive time to avoid idle timeouts on cloud load balancers
        ("grpc.keepalive_time_ms", 20000),
        ("grpc.keepalive_timeout_ms", 5000),
        ("grpc.http2.max_pings_without_data", 0),
        ("grpc.keepalive_permit_without_calls", 1),
    )
    options = dict(defaults)
    options.update(grpc_options or {})
    return tuple(options.items())


def configure_grpc_env():
    # disable informative logs by default, i.e.:
    # WARNING: All log messages before absl::InitializeLog() is called are written to STDERR
    # I0000 00:00:1739970744.889307   61962 ssl_transport_security.cc:1665] Handshake failed ...
    if os.environ.get("GRPC_VERBOSITY") is None:
        os.environ["GRPC_VERBOSITY"] = "ERROR"
    if os.environ.get("GLOG_minloglevel") is None:
        os.environ["GLOG_minloglevel"] = "2"


@contextmanager
def translate_grpc_exceptions():
    """Translate grpc exceptions to JumpstarterExceptions."""
    try:
        yield
    except grpc.aio.AioRpcError as e:
        if e.code().name == "UNAVAILABLE":
            # tls or other connection errors
            raise ConnectionError(f"grpc error: {e.details()}") from None
        if e.code().name == "UNKNOWN":
            # an error returned from our functions
            raise ConnectionError(f"grpc controller responded: {e.details()}") from None
        else:
            raise ConnectionError("grpc error") from e
    except grpc.RpcError as e:
        raise ConnectionError("grpc error") from e
    except ValueError as e:
        raise ConfigurationError("grpc error") from e
    except Exception as e:
        raise e



from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager, suppress
from pathlib import Path
from typing import Any, ClassVar, Literal, Optional, Self

import grpc
import yaml
from anyio.from_thread import start_blocking_portal
from pydantic import BaseModel, ConfigDict, Field, RootModel

from .common import ObjectMeta
from .grpc import call_credentials
from .tls import TLSConfigV1Alpha1
from jumpstarter.common.grpc import aio_secure_channel, ssl_channel_credentials
from jumpstarter.common.importlib import import_class
from jumpstarter.driver import Driver


class ExporterConfigV1Alpha1DriverInstanceProxy(BaseModel):
    ref: str


class ExporterConfigV1Alpha1DriverInstanceComposite(BaseModel):
    children: dict[str, ExporterConfigV1Alpha1DriverInstance] = Field(default_factory=dict)


class ExporterConfigV1Alpha1DriverInstanceBase(BaseModel):
    type: str
    config: dict[str, Any] = Field(default_factory=dict)
    children: dict[str, ExporterConfigV1Alpha1DriverInstance] = Field(default_factory=dict)


class ExporterConfigV1Alpha1DriverInstance(RootModel):
    root: (
        ExporterConfigV1Alpha1DriverInstanceBase
        | ExporterConfigV1Alpha1DriverInstanceComposite
        | ExporterConfigV1Alpha1DriverInstanceProxy
    )

    def instantiate(self) -> Driver:
        match self.root:
            case ExporterConfigV1Alpha1DriverInstanceBase():
                driver_class = import_class(self.root.type, [], True)

                children = {name: child.instantiate() for name, child in self.root.children.items()}

                return driver_class(children=children, **self.root.config)

            case ExporterConfigV1Alpha1DriverInstanceComposite():
                from jumpstarter_driver_composite.driver import Composite

                children = {name: child.instantiate() for name, child in self.root.children.items()}

                return Composite(children=children)

            case ExporterConfigV1Alpha1DriverInstanceProxy():
                from jumpstarter_driver_composite.driver import Proxy

                return Proxy(ref=self.root.ref)

    @classmethod
    def from_path(cls, path: str) -> ExporterConfigV1Alpha1DriverInstance:
        with open(path) as f:
            return cls.model_validate(yaml.safe_load(f))

    @classmethod
    def from_str(cls, config: str) -> ExporterConfigV1Alpha1DriverInstance:
        return cls.model_validate(yaml.safe_load(config))


class ExporterConfigV1Alpha1(BaseModel):
    BASE_PATH: ClassVar[Path] = Path("/etc/jumpstarter/exporters")

    alias: str = Field(default="default")

    apiVersion: Literal["jumpstarter.dev/v1alpha1"] = Field(default="jumpstarter.dev/v1alpha1")
    kind: Literal["ExporterConfig"] = Field(default="ExporterConfig")
    metadata: ObjectMeta

    endpoint: str
    tls: TLSConfigV1Alpha1 = Field(default_factory=TLSConfigV1Alpha1)
    token: str
    grpcOptions: dict[str, str | int] | None = Field(default_factory=dict)

    export: dict[str, ExporterConfigV1Alpha1DriverInstance] = Field(default_factory=dict)

    path: Path | None = Field(default=None)

    @classmethod
    def _get_path(cls, alias: str):
        return (cls.BASE_PATH / alias).with_suffix(".yaml")

    @classmethod
    def exists(cls, alias: str):
        return cls._get_path(alias).exists()

    @classmethod
    def load_path(cls, path: Path):
        with path.open() as f:
            config = cls.model_validate(yaml.safe_load(f))
            config.path = path
            return config

    @classmethod
    def load(cls, alias: str) -> Self:
        config = cls.load_path(cls._get_path(alias))
        config.alias = alias
        return config

    @classmethod
    def list(cls) -> list[Self]:
        exporters = []
        with suppress(FileNotFoundError):
            for entry in cls.BASE_PATH.iterdir():
                exporters.append(cls.load(entry.stem))
        return exporters

    @classmethod
    def dump_yaml(self, config: Self) -> str:
        return yaml.safe_dump(config.model_dump(mode="json", exclude={"alias", "path"}), sort_keys=False)

    @classmethod
    def save(cls, config: Self, path: Optional[str] = None) -> Path:
        # Set the config path before saving
        if path is None:
            config.path = cls._get_path(config.alias)
            config.path.parent.mkdir(parents=True, exist_ok=True)
        else:
            config.path = Path(path)
        with config.path.open(mode="w") as f:
            yaml.safe_dump(config.model_dump(mode="json", exclude={"alias", "path"}), f, sort_keys=False)
        return config.path

    @classmethod
    def delete(cls, alias: str) -> Path:
        path = cls._get_path(alias)
        path.unlink(missing_ok=True)
        return path

    @asynccontextmanager
    async def serve_unix_async(self):
        # dynamic import to avoid circular imports
        from jumpstarter.exporter import Session

        with Session(
            root_device=ExporterConfigV1Alpha1DriverInstance(children=self.export).instantiate(),
        ) as session:
            async with session.serve_unix_async() as path:
                yield path

    @contextmanager
    def serve_unix(self):
        with start_blocking_portal() as portal:
            with portal.wrap_async_context_manager(self.serve_unix_async()) as path:
                yield path

    async def serve(self):
        # dynamic import to avoid circular imports
        from jumpstarter.exporter import Exporter

        async def channel_factory():
            credentials = grpc.composite_channel_credentials(
                await ssl_channel_credentials(self.endpoint, self.tls),
                call_credentials("Exporter", self.metadata, self.token),
            )
            return aio_secure_channel(self.endpoint, credentials, self.grpcOptions)

        async with Exporter(
            channel_factory=channel_factory,
            device_factory=ExporterConfigV1Alpha1DriverInstance(children=self.export).instantiate,
            tls=self.tls,
            grpc_options=self.grpcOptions,
        ) as exporter:
            await exporter.serve()


class ExporterConfigListV1Alpha1(BaseModel):
    api_version: Literal["jumpstarter.dev/v1alpha1"] = Field(alias="apiVersion", default="jumpstarter.dev/v1alpha1")
    items: list[ExporterConfigV1Alpha1]
    kind: Literal["ExporterConfigList"] = Field(default="ExporterConfigList")

    def dump_json(self):
        return self.model_dump_json(indent=4, by_alias=True)

    def dump_yaml(self):
        return yaml.safe_dump(self.model_dump(mode="json", by_alias=True), indent=2)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)



# Ported from https://github.com/kubernetes/apimachinery/blob/v0.31.1/pkg/api/meta/conditions.go

from jumpstarter_protocol import kubernetes_pb2


def condition_present_and_equal(
    conditions: list[kubernetes_pb2.Condition], condition_type: str, status: str, reason: str | None = None
) -> bool:
    for condition in conditions:
        if condition.type == condition_type:
            if reason is None or condition.reason == reason:
                return condition.status == status
    return False


def condition_message(
    conditions: list[kubernetes_pb2.Condition], condition_type: str, reason: str | None = None
) -> str | None:
    for condition in conditions:
        if condition.type == condition_type:
            if reason is None or condition.reason == reason:
                return condition.message
    return None


def condition_true(conditions: list[kubernetes_pb2.Condition], condition_type: str) -> bool:
    return condition_present_and_equal(conditions, condition_type, "True")


def condition_false(conditions: list[kubernetes_pb2.Condition], condition_type: str) -> bool:
    return condition_present_and_equal(conditions, condition_type, "False")



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from ipaddress import IPv4Address
from typing import Tuple
from uuid import uuid4

import grpc
import pytest
from anyio import Event, create_memory_object_stream
from anyio.abc import AnyByteStream
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from jumpstarter_protocol import (
    jumpstarter_pb2,
    jumpstarter_pb2_grpc,
    router_pb2_grpc,
)

from jumpstarter.streams.common import forward_stream
from jumpstarter.streams.router import RouterStream


@dataclass(kw_only=True)
class MockRouter(router_pb2_grpc.RouterServiceServicer):
    pending: dict[str, AnyByteStream] = field(default_factory=dict)

    async def Stream(self, _request_iterator, context):
        event = Event()
        context.add_done_callback(lambda _: event.set())
        authorization = dict(list(context.invocation_metadata()))["authorization"]
        async with RouterStream(context=context) as stream:
            if authorization in self.pending:
                async with forward_stream(stream, self.pending[authorization]):
                    await event.wait()
            else:
                self.pending[authorization] = stream
                await event.wait()
                del self.pending[authorization]


@dataclass(kw_only=True)
class MockController(jumpstarter_pb2_grpc.ControllerServiceServicer):
    router_endpoint: str
    status: Tuple[
        MemoryObjectSendStream[jumpstarter_pb2.StatusResponse],
        MemoryObjectReceiveStream[jumpstarter_pb2.StatusResponse],
    ] = field(init=False, default_factory=lambda: create_memory_object_stream[jumpstarter_pb2.StatusResponse](32))
    queue: Tuple[MemoryObjectSendStream[str], MemoryObjectReceiveStream[str]] = field(
        init=False, default_factory=lambda: create_memory_object_stream[str](32)
    )
    leases: dict[str, int | str] = field(init=False, default_factory=dict)

    async def Register(self, request, context):
        return jumpstarter_pb2.RegisterResponse(uuid=str(uuid4()))

    async def Unregister(self, request, context):
        return jumpstarter_pb2.UnregisterResponse()

    async def Dial(self, request, context):
        token = str(uuid4())
        await self.queue[0].send(token)
        return jumpstarter_pb2.DialResponse(router_endpoint=self.router_endpoint, router_token=token)

    async def Status(self, request, context):
        async for status in self.status[1]:
            yield status

    async def Listen(self, request, context):
        async for token in self.queue[1]:
            yield jumpstarter_pb2.ListenResponse(router_endpoint=self.router_endpoint, router_token=token)


@pytest.fixture
def anyio_backend():
    return "asyncio"


key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

cert = (
    x509.CertificateBuilder()
    .subject_name(x509.Name([]))
    .issuer_name(x509.Name([]))
    .public_key(key.public_key())
    .serial_number(x509.random_serial_number())
    .not_valid_before(datetime.now())
    .not_valid_after(datetime.now() + timedelta(days=365))
    .add_extension(x509.SubjectAlternativeName([x509.IPAddress(IPv4Address("127.0.0.1"))]), critical=False)
    .sign(private_key=key, algorithm=hashes.SHA256(), backend=default_backend())
)

tls_crt = cert.public_bytes(serialization.Encoding.PEM)
tls_key = key.private_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PrivateFormat.TraditionalOpenSSL,
    encryption_algorithm=serialization.NoEncryption(),
)


@pytest.fixture
async def mock_controller(tmp_path, monkeypatch):
    monkeypatch.setenv("JUMPSTARTER_GRPC_INSECURE", "1")

    server = grpc.aio.server()
    port = server.add_secure_port(
        "127.0.0.1:0", grpc.ssl_server_credentials(private_key_certificate_chain_pairs=[(tls_key, tls_crt)])
    )

    controller = MockController(router_endpoint=f"127.0.0.1:{port}")
    router = MockRouter()

    jumpstarter_pb2_grpc.add_ControllerServiceServicer_to_server(controller, server)
    router_pb2_grpc.add_RouterServiceServicer_to_server(router, server)

    await server.start()
    try:
        yield f"127.0.0.1:{port}"
    finally:
        await server.stop(grace=None)


pytest_plugins = ["pytester"]



import logging
from dataclasses import dataclass, field

from anyio.abc import ObjectStream
from tqdm import tqdm

TQDM_KWARGS = {
    "unit": "B",
    "unit_scale": True,
    "unit_divisor": 1024,
}

LOGGER = logging.getLogger(__name__)


# Copied from https://github.com/tqdm/tqdm/pull/1172
class logging_tqdm(tqdm):
    """
    A version of tqdm that outputs the progress bar
    to Python logging instead of the console.
    The progress will be logged with the info level.
    Parameters
    ----------
    logger   : logging.Logger, optional
      Which logger to output to (default: logger.getLogger('tqdm.contrib.logging')).
    All other parameters are passed on to regular tqdm,
    with the following changed default:
    mininterval: 1
    bar_format: '{desc}{percentage:3.0f}%{r_bar}'
    desc: 'progress: '
    Example
    -------
    ```python
    import logging
    from time import sleep
    from tqdm.contrib.logging import logging_tqdm
    LOG = logging.getLogger(__name__)
    if __name__ == '__main__':
        logging.basicConfig(level=logging.INFO)
        for _ in logging_tqdm(range(10), mininterval=1, logger=LOG):
            sleep(0.3)  # assume processing one item takes less than mininterval
    ```
    """

    def __init__(
        self,
        *args,
        # logger=None,  # type: logging.Logger
        # mininterval=1,  # type: float
        # bar_format='{desc}{percentage:3.0f}%{r_bar}',  # type: str
        # desc='progress: ',  # type: str
        **kwargs,
    ):
        if len(args) >= 2:
            # Note: Due to Python 2 compatibility, we can't declare additional
            #   keyword arguments in the signature.
            #   As a result, we could get (due to the defaults below):
            #     TypeError: __init__() got multiple values for argument 'desc'
            #   This will raise a more descriptive error message.
            #   Calling dummy init to avoid attribute errors when __del__ is called
            super(logging_tqdm, self).__init__([], disable=True)
            raise ValueError("only iterable may be used as a positional argument")
        tqdm_kwargs = kwargs.copy()
        self._logger = tqdm_kwargs.pop("logger", None)
        tqdm_kwargs.setdefault("mininterval", 1)
        tqdm_kwargs.setdefault("bar_format", "{desc}{percentage:3.0f}%{r_bar}")
        tqdm_kwargs.setdefault("desc", "progress: ")
        self._last_log_n = -1
        super(logging_tqdm, self).__init__(*args, **tqdm_kwargs)

    def _get_logger(self):
        if self._logger is not None:
            return self._logger
        return LOGGER

    def display(self, msg=None, pos=None):
        if not self.n:
            # skip progress bar before having processed anything
            LOGGER.debug("ignoring message before any progress: %r", self.n)
            return
        if self.n == self._last_log_n:
            # avoid logging for the same progress multiple times
            LOGGER.debug("ignoring log message with same n: %r", self.n)
            return
        self._last_log_n = self.n
        if msg is None:
            msg = self.__str__()
        if not msg:
            LOGGER.debug("ignoring empty message: %r", msg)
            return
        self._get_logger().info("%s", msg)


@dataclass(kw_only=True)
class ProgressStream(ObjectStream[bytes]):
    stream: ObjectStream
    logging: bool = False

    __recv: tqdm = field(init=False, default=None)
    __send: tqdm = field(init=False, default=None)

    def __del__(self):
        if self.__recv is not None:
            self.__recv.close()
        if self.__send is not None:
            self.__send.close()

    async def receive(self):
        item = await self.stream.receive()

        if self.__recv is None:
            if self.logging:
                self.__recv = logging_tqdm(desc="transfer", **TQDM_KWARGS)
            else:
                self.__recv = tqdm(desc="transfer", **TQDM_KWARGS)

        self.__recv.update(len(item))

        return item

    async def send(self, item):
        if self.__send is None:
            if self.logging:
                self.__send = logging_tqdm(desc="transfer", **TQDM_KWARGS)
            else:
                self.__send = tqdm(desc="transfer", **TQDM_KWARGS)

        self.__send.update(len(item))

        await self.stream.send(item)

    async def send_eof(self):
        await self.stream.send_eof()

    async def aclose(self):
        await self.stream.aclose()



from dataclasses import dataclass

from aiohttp import ClientError, StreamReader
from anyio import BrokenResourceError, EndOfStream
from anyio.abc import ObjectStream


@dataclass(frozen=True, kw_only=True, slots=True)
class AiohttpStreamReaderStream(ObjectStream[bytes]):
    reader: StreamReader

    async def send(self, item: bytes):
        raise BrokenResourceError

    async def receive(self) -> bytes:
        try:
            item = await self.reader.readany()
        except ClientError as e:
            raise BrokenResourceError from e
        if len(item) == 0:
            raise EndOfStream
        return item

    async def send_eof(self):
        pass

    async def aclose(self):
        pass



from typing import Annotated, Literal, Union
from uuid import UUID

from pydantic import BaseModel, Field, Json


class ClientStreamResource(BaseModel):
    kind: Literal["client_stream"] = "client_stream"
    uuid: UUID


class PresignedRequestResource(BaseModel):
    kind: Literal["presigned_request"] = "presigned_request"
    headers: dict[str, str]
    url: str
    method: Literal["GET", "PUT"]


Resource = Annotated[
    Union[ClientStreamResource, PresignedRequestResource],
    Field(discriminator="kind"),
]


class ResourceMetadata(BaseModel):
    resource: Json[Resource]



from .metadata import Metadata
from .tempfile import TemporarySocket, TemporaryTcpListener, TemporaryUnixListener

__all__ = ["Metadata", "TemporarySocket", "TemporaryUnixListener", "TemporaryTcpListener"]



import pytest

from . import export
from .decorators import (
    MARKER_DRIVERCALL,
    MARKER_MAGIC,
    MARKER_STREAMING_DRIVERCALL,
)


class Functions:
    @export
    def function(self):
        pass

    @export
    async def asyncfunction(self):
        pass

    @export
    def generator(self):
        yield

    @export
    async def asyncgenerator(self):
        yield


def test_driver_decorators():
    functions = Functions()

    assert getattr(functions.function, MARKER_DRIVERCALL) == MARKER_MAGIC
    assert getattr(functions.asyncfunction, MARKER_DRIVERCALL) == MARKER_MAGIC
    assert getattr(functions.generator, MARKER_STREAMING_DRIVERCALL) == MARKER_MAGIC
    assert getattr(functions.asyncgenerator, MARKER_STREAMING_DRIVERCALL) == MARKER_MAGIC

    with pytest.raises(ValueError):
        export(None)



import grpc

from .common import ObjectMeta


def call_credentials(kind: str, metadata: ObjectMeta, token: str):
    def metadata_call_credentials(context: grpc.AuthMetadataContext, callback: grpc.AuthMetadataPluginCallback):
        callback(
            [
                ("jumpstarter-kind", kind),
                ("jumpstarter-namespace", metadata.namespace),
                ("jumpstarter-name", metadata.name),
            ],
            None,
        )

    return grpc.composite_call_credentials(
        grpc.metadata_call_credentials(metadata_call_credentials),
        grpc.access_token_call_credentials(token),
    )



from contextlib import asynccontextmanager
from typing import Annotated, Literal, Union
from uuid import UUID

import grpc
from jumpstarter_protocol.jumpstarter.v1 import router_pb2_grpc
from pydantic import BaseModel, Field, Json

from jumpstarter.common.grpc import aio_secure_channel, ssl_channel_credentials
from jumpstarter.streams.common import forward_stream
from jumpstarter.streams.router import RouterStream


class ResourceStreamRequest(BaseModel):
    kind: Literal["resource"] = "resource"
    uuid: UUID


class DriverStreamRequest(BaseModel):
    kind: Literal["driver"] = "driver"
    uuid: UUID
    method: str


StreamRequest = Annotated[
    Union[ResourceStreamRequest, DriverStreamRequest],
    Field(discriminator="kind"),
]


class StreamRequestMetadata(BaseModel):
    request: Json[StreamRequest]


@asynccontextmanager
async def connect_router_stream(endpoint, token, stream, tls_config, grpc_options):
    credentials = grpc.composite_channel_credentials(
        await ssl_channel_credentials(endpoint, tls_config),
        grpc.access_token_call_credentials(token),
    )

    async with aio_secure_channel(endpoint, credentials, grpc_options) as channel:
        router = router_pb2_grpc.RouterServiceStub(channel)
        context = router.Stream(metadata=())
        async with RouterStream(context=context) as s:
            async with forward_stream(s, stream):
                yield



import os
from pathlib import Path
from typing import Annotated, ClassVar, Literal, Optional, Self

import yaml
from pydantic import BaseModel, ConfigDict, Field, PlainSerializer
from pydantic.functional_validators import BeforeValidator

from .client import ClientConfigV1Alpha1
from .common import CONFIG_PATH


def _serialize_current_client(v: ClientConfigV1Alpha1 | None) -> str | None:
    if v:
        return v.alias
    else:
        return None


def _validate_current_client(v: str | ClientConfigV1Alpha1 | None) -> ClientConfigV1Alpha1 | None:
    match v:
        case str():
            try:
                return ClientConfigV1Alpha1.load(v)
            except FileNotFoundError:
                return None
        case ClientConfigV1Alpha1():
            return v
        case None:
            return None
        case _:
            raise ValueError("current-client is not of type str | ClientConfigV1Alpha1 | None")


CurrentClient = Annotated[
    ClientConfigV1Alpha1 | None,
    PlainSerializer(_serialize_current_client),
    BeforeValidator(_validate_current_client),
]


class UserConfigV1Alpha1Config(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    current_client: CurrentClient | None = Field(alias="current-client", default=None)


class UserConfigV1Alpha1(BaseModel):
    """The user configuration for the Jumpstarter CLI."""

    BASE_CONFIG_PATH: ClassVar[Path] = CONFIG_PATH
    USER_CONFIG_PATH: ClassVar[Path] = CONFIG_PATH / "config.yaml"

    apiVersion: Literal["jumpstarter.dev/v1alpha1"] = Field(default="jumpstarter.dev/v1alpha1")
    kind: Literal["UserConfig"] = Field(default="UserConfig")
    config: UserConfigV1Alpha1Config

    @classmethod
    def exists(cls) -> bool:
        """Check if a user config exists."""
        return os.path.exists(cls.USER_CONFIG_PATH)

    @classmethod
    def load(cls) -> Self:
        """Load the user config from the default path."""
        if cls.exists() is False:
            raise FileNotFoundError(f"User config file not found at '{cls.USER_CONFIG_PATH}'.")

        with open(cls.USER_CONFIG_PATH) as f:
            config: dict = yaml.safe_load(f)
            return cls.model_validate(config)

    @classmethod
    def load_or_create(cls) -> Self:
        """Check if a user config exists, otherwise create an empty one."""
        if cls.exists() is False:
            os.makedirs(cls.BASE_CONFIG_PATH, exist_ok=True)
            config = cls(config=UserConfigV1Alpha1Config(current_client=None))
            cls.save(config)
            return config
        # Always return the current user config if it exists
        return cls.load()

    @classmethod
    def save(cls, config: Self, path: Optional[str] = None) -> Path:
        """Save a user config as YAML."""

        with open(path or cls.USER_CONFIG_PATH, "w") as f:
            yaml.safe_dump(config.model_dump(mode="json", by_alias=True), f, sort_keys=False)
        return path or cls.USER_CONFIG_PATH

    def use_client(self, name: Optional[str]) -> Path | None:
        """Updates the current client and saves the user config."""
        if name is not None:
            self.config.current_client = ClientConfigV1Alpha1.load(name)
        else:
            self.config.current_client = None
        self.save(self)
        if self.config.current_client is not None:
            return self.config.current_client.path
        else:
            return None



import errno
import os
from logging import Logger
from typing import Literal

from anyio import fail_after, sleep
from anyio.abc import AnyByteStream
from anyio.streams.file import FileReadStream, FileWriteStream


async def wait_for_storage_device(  # noqa: C901
    storage_device: os.PathLike,
    mode: Literal["wb", "rb"],
    timeout: int = 10,
    *,
    logger: Logger | None = None,
) -> os.PathLike:
    with fail_after(timeout):
        while True:
            # https://stackoverflow.com/a/2774125
            try:
                match mode:
                    case "wb":
                        fd = os.open(storage_device, os.O_WRONLY)
                    case "rb":
                        fd = os.open(storage_device, os.O_RDONLY)
                    case _:
                        raise ValueError("invalid mode: {}".format(mode))
                with os.fdopen(fd, mode):  # to prevent fd from leaking
                    if os.lseek(fd, 0, os.SEEK_END) > 0:
                        if logger:
                            logger.info("storage device {} is ready".format(storage_device))
                        break
                if logger:
                    logger.debug("waiting for storage device {} to have a nonzero size".format(storage_device))
            except FileNotFoundError:
                if logger:
                    logger.debug("waiting for storage device {} to appear".format(storage_device))
            except OSError as e:
                match e.errno:
                    case errno.ENOMEDIUM | errno.EIO:
                        if logger:
                            logger.debug("waiting for storage device {} to be ready".format(storage_device))
                    case _:
                        raise

            await sleep(1)

    return storage_device


async def write_to_storage_device(
    storage_device: os.PathLike,
    resource: AnyByteStream,
    timeout: int = 10,
    fsync_timeout: int = 900,
    leeway: int = 6,
    *,
    logger: Logger | None = None,
):
    path = await wait_for_storage_device(
        storage_device,
        mode="wb",
        timeout=timeout,
        logger=logger,
    )
    with os.fdopen(os.open(path, os.O_WRONLY), "wb") as file:
        async with FileWriteStream(file) as stream:
            total_bytes = 0
            next_print = 0
            async for chunk in resource:
                await stream.send(chunk)
                if logger:
                    total_bytes += len(chunk)
                    if total_bytes > next_print:
                        logger.info(
                            "written {} MB to storage device {}".format(
                                total_bytes / (1024 * 1024),
                                storage_device,
                            )
                        )
                        next_print += 50 * 1024 * 1024

            with fail_after(fsync_timeout):
                while True:
                    try:
                        if logger:
                            logger.info("fsyncing storage device {}, please wait".format(storage_device))
                        os.fsync(file.fileno())
                    except OSError as e:
                        if e.errno == errno.EIO:
                            await sleep(1)
                            continue
                        else:
                            raise
                    else:
                        break

            await sleep(leeway)


async def read_from_storage_device(
    storage_device: os.PathLike,
    resource: AnyByteStream,
    timeout: int = 10,
    *,
    logger: Logger | None = None,
):
    path = await wait_for_storage_device(
        storage_device,
        mode="rb",
        timeout=timeout,
        logger=logger,
    )
    with os.fdopen(os.open(path, os.O_RDONLY), "rb") as file:
        async with FileReadStream(file) as stream:
            total_bytes = 0
            next_print = 0
            async for chunk in stream:
                await resource.send(chunk)
                if logger:
                    total_bytes += len(chunk)
                    if total_bytes > next_print:
                        logger.info(
                            "read {} MB from storage device {}".format(
                                total_bytes / (1024 * 1024),
                                storage_device,
                            )
                        )
                        next_print += 50 * 1024 * 1024



import logging
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    ExitStack,
    asynccontextmanager,
    contextmanager,
)
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any

from anyio import create_task_group, fail_after, sleep
from anyio.from_thread import BlockingPortal
from grpc.aio import Channel
from jumpstarter_protocol import jumpstarter_pb2, jumpstarter_pb2_grpc

from .exceptions import LeaseError
from jumpstarter.client import client_from_path
from jumpstarter.client.grpc import ClientService
from jumpstarter.common import TemporaryUnixListener
from jumpstarter.common.condition import condition_false, condition_message, condition_present_and_equal, condition_true
from jumpstarter.common.grpc import translate_grpc_exceptions
from jumpstarter.common.streams import connect_router_stream
from jumpstarter.config.tls import TLSConfigV1Alpha1

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class Lease(AbstractContextManager, AbstractAsyncContextManager):
    channel: Channel
    duration: timedelta
    selector: str
    portal: BlockingPortal
    namespace: str
    name: str | None = field(default=None)
    allow: list[str]
    unsafe: bool
    release: bool = True  # release on contexts exit
    controller: jumpstarter_pb2_grpc.ControllerServiceStub = field(init=False)
    tls_config: TLSConfigV1Alpha1 = field(default_factory=TLSConfigV1Alpha1)
    grpc_options: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()

        self.controller = jumpstarter_pb2_grpc.ControllerServiceStub(self.channel)
        self.svc = ClientService(channel=self.channel, namespace=self.namespace)
        self.manager = self.portal.wrap_async_context_manager(self)

    async def _create(self):
        logger.debug("Creating lease request for selector %s for duration %s", self.selector, self.duration)
        with translate_grpc_exceptions():
            self.name = (
                await self.svc.CreateLease(
                    selector=self.selector,
                    duration=self.duration,
                )
            ).name
        logger.info("Created lease request for selector %s for duration %s", self.selector, self.duration)

    async def get(self):
        with translate_grpc_exceptions():
            svc = ClientService(channel=self.channel, namespace=self.namespace)
            return await svc.GetLease(name=self.name)

    def request(self):
        """Request a lease, or verifies a lease which was already created.

        :return: lease
        :rtype: Lease
        :raises LeaseError: if lease is unsatisfiable
        :raises LeaseError: if lease is not pending
        :raises TimeoutError: if lease is not ready after timeout
        """
        return self.portal.call(self.request_async)

    async def request_async(self):
        """Request a lease, or verifies a lease which was already created.

        :return: lease
        :rtype: Lease
        :raises LeaseError: if lease is unsatisfiable
        :raises LeaseError: if lease is not pending
        :raises TimeoutError: if lease is not ready after timeout
        """
        if self.name:
            logger.debug("Using existing lease %s", self.name)
        else:
            await self._create()
        return await self._acquire()

    async def _acquire(self):
        """Acquire a lease.

        Makes sure the lease is ready, and returns the lease object.
        """
        with fail_after(300):  # TODO: configurable timeout
            while True:
                logger.debug("Polling Lease %s", self.name)
                result = await self.get()
                # lease ready
                if condition_true(result.conditions, "Ready"):
                    logger.debug("Lease %s acquired", self.name)
                    return self
                # lease unsatisfiable
                if condition_true(result.conditions, "Unsatisfiable"):
                    message = condition_message(result.conditions, "Unsatisfiable")
                    logger.debug(
                        "Lease %s cannot be satisfied: %s",
                        self.name,
                        condition_message(result.conditions, "Unsatisfiable"),
                    )
                    raise LeaseError(f"the lease cannot be satisfied: {message}")

                # lease not pending
                if condition_false(result.conditions, "Pending"):
                    raise LeaseError(
                        f"Lease {self.name} is not in pending, but it isn't in Ready or Unsatisfiable state either"
                    )

                # lease released
                if condition_present_and_equal(result.conditions, "Ready", "False", "Released"):
                    raise LeaseError(f"lease {self.name} released")

                await sleep(1)

    async def __aenter__(self):
        return await self.request_async()

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self.release:
            logger.info("Releasing Lease %s", self.name)
            await self.svc.DeleteLease(
                name=self.name,
            )

    def __enter__(self):
        # wraps the async context manager enter
        return self.manager.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        # wraps the async context manager exit
        return self.manager.__exit__(exc_type, exc_value, traceback)

    async def handle_async(self, stream):
        logger.debug("Connecting to Lease with name %s", self.name)
        response = await self.controller.Dial(jumpstarter_pb2.DialRequest(lease_name=self.name))
        async with connect_router_stream(
            response.router_endpoint, response.router_token, stream, self.tls_config, self.grpc_options
        ):
            pass

    @asynccontextmanager
    async def serve_unix_async(self):
        async with TemporaryUnixListener(self.handle_async) as path:
            yield path

    @asynccontextmanager
    async def monitor_async(self, threshold: timedelta = timedelta(minutes=5)):
        async def _monitor():
            while True:
                lease = await self.get()
                # TODO: use effective_end_time as the authoritative source for lease end time
                if lease.effective_begin_time:
                    end_time = lease.effective_begin_time + lease.duration
                    remain = end_time - datetime.now(tz=datetime.now().astimezone().tzinfo)
                    if remain < timedelta(0):
                        # lease already expired, stopping monitor
                        logger.info("Lease {} ended at {}".format(self.name, end_time))
                        break
                    elif remain < threshold:
                        # lease expiring soon, check again on expected expiration time in case it's extended
                        logger.info("Lease {} ending soon in {} at {}".format(self.name, remain, end_time))
                        await sleep(threshold.total_seconds())
                    else:
                        # lease still active, check again in 5 seconds
                        await sleep(5)
                else:
                    await sleep(1)

        async with create_task_group() as tg:
            tg.start_soon(_monitor)
            try:
                yield
            finally:
                tg.cancel_scope.cancel()

    @asynccontextmanager
    async def connect_async(self, stack):
        async with self.serve_unix_async() as path:
            async with client_from_path(path, self.portal, stack, allow=self.allow, unsafe=self.unsafe) as client:
                yield client

    @contextmanager
    def connect(self):
        with ExitStack() as stack:
            with self.portal.wrap_async_context_manager(self.connect_async(stack)) as client:
                yield client

    @contextmanager
    def serve_unix(self):
        with self.portal.wrap_async_context_manager(self.serve_unix_async()) as path:
            yield path

    @contextmanager
    def monitor(self, threshold: timedelta = timedelta(minutes=5)):
        with self.portal.wrap_async_context_manager(self.monitor_async(threshold)):
            yield



import sys


class JumpstarterException(Exception):
    """Base class for jumpstarter-specific errors.

    This class should not be raised directly, but should be used as a base
    class for all jumpstarter-specific errors.
    It handles the __cause__ attribute so the jumpstarter errors could be raised as

    .. code-block:: python

        raise SomeError("message") from original_exception
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def __str__(self):
        if self.__cause__:
            return f"{self.message} (Caused by: {self.__cause__})"
        return f"{self.message}"

    def print(self, message: str | None = None):
        ANSI_RED = "\033[91m"
        ANSI_CLEAR = "\033[0m"
        print(f"{ANSI_RED}{self}{ANSI_CLEAR}", file=sys.stderr)


class ConnectionError(JumpstarterException):
    """Raised when a connection to a jumpstarter server fails."""

    pass


class ConfigurationError(JumpstarterException):
    """Raised when a configuration error exists."""

    pass


class ArgumentError(JumpstarterException):
    """Raised when a cli argument is not valid."""

    pass


class FileAccessError(JumpstarterException):
    """Raised when a file access error occurs."""

    pass


class FileNotFoundError(JumpstarterException, FileNotFoundError):
    """Raised when a file is not found."""

    pass



import logging
from collections import deque

from jumpstarter_protocol import jumpstarter_pb2


class LogHandler(logging.Handler):
    def __init__(self, queue: deque):
        logging.Handler.__init__(self)
        self.queue = queue
        self.listener = None

    def enqueue(self, record):
        self.queue.append(record)

    def prepare(self, record):
        return jumpstarter_pb2.LogStreamResponse(
            uuid="",
            severity=record.levelname,
            message=self.format(record),
        )

    def emit(self, record):
        try:
            self.enqueue(self.prepare(record))
        except Exception:
            self.handleError(record)



from contextlib import asynccontextmanager, contextmanager, nullcontext
from os import PathLike
from pathlib import Path
from socket import AddressFamily
from tempfile import TemporaryDirectory

from anyio import create_task_group, create_tcp_listener, create_unix_listener
from anyio.abc import SocketAttribute
from xdg_base_dirs import xdg_runtime_dir


@contextmanager
def TemporarySocket():
    with TemporaryDirectory(dir=xdg_runtime_dir(), prefix="jumpstarter-") as tempdir:
        yield Path(tempdir) / "socket"


@asynccontextmanager
async def TemporaryUnixListener(handler, path: PathLike | None = None):
    if path is not None:
        cm = nullcontext(path)
    else:
        cm = TemporarySocket()

    with cm as path:
        async with await create_unix_listener(path) as listener:
            async with create_task_group() as tg:
                tg.start_soon(listener.serve, handler, tg)
                try:
                    yield path
                finally:
                    tg.cancel_scope.cancel()


@asynccontextmanager
async def TemporaryTcpListener(
    handler, local_host="127.0.0.1", local_port=0, family=AddressFamily.AF_UNSPEC, backlog=65536, reuse_port=True
):
    async with await create_tcp_listener(
        local_host=local_host,
        local_port=local_port,
        family=family,
        backlog=backlog,
        reuse_port=reuse_port,
    ) as listener:
        async with create_task_group() as tg:
            tg.start_soon(listener.serve, handler, tg)
            try:
                yield listener.extra(SocketAttribute.local_address)
            finally:
                tg.cancel_scope.cancel()



from typing import Any

from google.protobuf import json_format, struct_pb2
from pydantic import TypeAdapter

adapter = TypeAdapter(Any)


def encode_value(v: Any):
    return json_format.ParseDict(adapter.dump_python(v, mode="json"), struct_pb2.Value())


def decode_value(v: struct_pb2.Value) -> Any:
    return adapter.validate_python(json_format.MessageToDict(v))



import os
from contextlib import ExitStack, asynccontextmanager, contextmanager

from anyio.from_thread import start_blocking_portal

from jumpstarter.client import client_from_path
from jumpstarter.config.client import _allow_from_env
from jumpstarter.config.env import JUMPSTARTER_HOST


@asynccontextmanager
async def env_async(portal, stack):
    """Provide a client for an existing JUMPSTARTER_HOST environment variable.

    Async version of env()

    This is useful when interacting with an already established Jumpstarter shell,
    to either a local exporter or a remote one.
    """
    host = os.environ.get(JUMPSTARTER_HOST, None)
    if host is None:
        raise RuntimeError(f"{JUMPSTARTER_HOST} not set")

    allow, unsafe = _allow_from_env()

    async with client_from_path(host, portal, stack, allow=allow, unsafe=unsafe) as client:
        try:
            yield client
        finally:
            if hasattr(client, "close"):
                client.close()


@contextmanager
def env():
    """Provide a client for an existing JUMPSTARTER_HOST environment variable.

    This is useful when interacting with an already established Jumpstarter shell,
    to either a local exporter or a remote one.
    """
    with start_blocking_portal() as portal:
        with ExitStack() as stack:
            with portal.wrap_async_context_manager(env_async(portal, stack)) as client:
                yield client



import logging
from collections import deque
from contextlib import AbstractContextManager, asynccontextmanager, contextmanager, suppress
from dataclasses import dataclass, field
from logging.handlers import QueueHandler
from uuid import UUID

import grpc
from anyio import Event, TypedAttributeLookupError, sleep
from anyio.from_thread import start_blocking_portal
from jumpstarter_protocol import (
    jumpstarter_pb2,
    jumpstarter_pb2_grpc,
    router_pb2_grpc,
)

from .logging import LogHandler
from jumpstarter.common import Metadata, TemporarySocket
from jumpstarter.common.streams import StreamRequestMetadata
from jumpstarter.driver import Driver
from jumpstarter.streams.common import forward_stream
from jumpstarter.streams.metadata import MetadataStreamAttributes
from jumpstarter.streams.router import RouterStream

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class Session(
    jumpstarter_pb2_grpc.ExporterServiceServicer,
    router_pb2_grpc.RouterServiceServicer,
    Metadata,
    AbstractContextManager,
):
    root_device: Driver
    mapping: dict[UUID, Driver]

    _logging_queue: deque = field(init=False)
    _logging_handler: QueueHandler = field(init=False)

    def __enter__(self):
        logging.getLogger().addHandler(self._logging_handler)
        self.root_device.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.root_device.close()
        logging.getLogger().removeHandler(self._logging_handler)

    def __init__(self, *args, root_device, **kwargs):
        super().__init__(*args, **kwargs)

        self.root_device = root_device
        self.mapping = {u: i for (u, _, _, i) in self.root_device.enumerate()}

        self._logging_queue = deque(maxlen=32)
        self._logging_handler = LogHandler(self._logging_queue)

    @asynccontextmanager
    async def serve_port_async(self, port):
        server = grpc.aio.server()
        server.add_insecure_port(port)

        jumpstarter_pb2_grpc.add_ExporterServiceServicer_to_server(self, server)
        router_pb2_grpc.add_RouterServiceServicer_to_server(self, server)

        await server.start()
        try:
            yield
        finally:
            await server.stop(grace=None)

    @asynccontextmanager
    async def serve_unix_async(self):
        with TemporarySocket() as path:
            async with self.serve_port_async(f"unix://{path}"):
                yield path

    @contextmanager
    def serve_unix(self):
        with start_blocking_portal() as portal:
            with portal.wrap_async_context_manager(self.serve_unix_async()) as path:
                yield path

    def __getitem__(self, key: UUID):
        return self.mapping[key]

    async def GetReport(self, request, context):
        logger.debug("GetReport()")
        return jumpstarter_pb2.GetReportResponse(
            uuid=str(self.uuid),
            labels=self.labels,
            reports=[
                instance.report(parent=parent, name=name)
                for (_, parent, name, instance) in self.root_device.enumerate()
            ],
        )

    async def DriverCall(self, request, context):
        logger.debug("DriverCall(uuid=%s, method=%s)", request.uuid, request.method)
        return await self[UUID(request.uuid)].DriverCall(request, context)

    async def StreamingDriverCall(self, request, context):
        logger.debug("StreamingDriverCall(uuid=%s, method=%s)", request.uuid, request.method)
        async for v in self[UUID(request.uuid)].StreamingDriverCall(request, context):
            yield v

    async def Stream(self, _request_iterator, context):
        request = StreamRequestMetadata(**dict(list(context.invocation_metadata()))).request
        logger.debug("Streaming(%s)", request)
        async with self[request.uuid].Stream(request, context) as stream:
            metadata = []
            with suppress(TypedAttributeLookupError):
                metadata.extend(stream.extra(MetadataStreamAttributes.metadata).items())
            await context.send_initial_metadata(metadata)

            async with RouterStream(context=context) as remote:
                async with forward_stream(remote, stream):
                    event = Event()
                    context.add_done_callback(lambda _: event.set())
                    await event.wait()

    async def LogStream(self, request, context):
        while True:
            try:
                yield self._logging_queue.popleft()
            except IndexError:
                await sleep(0.5)



import pytest

from .importlib import import_class


def test_import_class():
    import_class("os.open", [], True)

    with pytest.raises(ImportError):
        import_class("os.invalid", [], True)

    with pytest.raises(ImportError):
        import_class("os.open", [], False)

    import_class("os.open", ["os.*"], False)

    with pytest.raises(ImportError):
        import_class("os.open", ["sys.*"], False)

    with pytest.raises(ImportError):
        import_class("os", [], True)



from collections import OrderedDict, defaultdict
from contextlib import ExitStack, asynccontextmanager
from graphlib import TopologicalSorter
from uuid import UUID

import grpc
from anyio.from_thread import BlockingPortal
from google.protobuf import empty_pb2
from jumpstarter_protocol import jumpstarter_pb2_grpc

from jumpstarter.client import DriverClient
from jumpstarter.common.importlib import import_class


@asynccontextmanager
async def client_from_path(path: str, portal: BlockingPortal, stack: ExitStack, allow: list[str], unsafe: bool):
    async with grpc.aio.secure_channel(
        f"unix://{path}", grpc.local_channel_credentials(grpc.LocalConnectionType.UDS)
    ) as channel:
        yield await client_from_channel(channel, portal, stack, allow, unsafe)


async def client_from_channel(
    channel: grpc.aio.Channel,
    portal: BlockingPortal,
    stack: ExitStack,
    allow: list[str],
    unsafe: bool,
) -> DriverClient:
    topo = defaultdict(list)
    last_seen = {}
    reports = {}
    clients = OrderedDict()

    response = await jumpstarter_pb2_grpc.ExporterServiceStub(channel).GetReport(empty_pb2.Empty())

    for index, report in enumerate(response.reports):
        topo[index] = []

        last_seen[report.uuid] = index

        if report.parent_uuid != "":
            parent_index = last_seen[report.parent_uuid]
            topo[parent_index].append(index)

        reports[index] = report

    for index in TopologicalSorter(topo).static_order():
        report = reports[index]

        client_class = import_class(report.labels["jumpstarter.dev/client"], allow, unsafe)
        client = client_class(
            uuid=UUID(report.uuid),
            labels=report.labels,
            channel=channel,
            portal=portal,
            stack=stack.enter_context(ExitStack()),
            children={reports[k].labels["jumpstarter.dev/name"]: clients[k] for k in topo[index]},
        )

        clients[index] = client

    return clients.popitem(last=True)[1]



from dataclasses import field
from uuid import UUID, uuid4

from pydantic.dataclasses import dataclass


@dataclass(kw_only=True, slots=True)
class Metadata:
    uuid: UUID = field(default_factory=uuid4)
    labels: dict[str, str] = field(default_factory=dict)

    @property
    def name(self):
        return self.labels.get("jumpstarter.dev/name", "unknown")



import logging
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from dataclasses import dataclass, field

import grpc
from anyio import connect_unix, create_memory_object_stream, create_task_group, sleep
from google.protobuf import empty_pb2
from jumpstarter_protocol import (
    jumpstarter_pb2,
    jumpstarter_pb2_grpc,
)

from jumpstarter.common import Metadata
from jumpstarter.common.streams import connect_router_stream
from jumpstarter.config.tls import TLSConfigV1Alpha1
from jumpstarter.driver import Driver
from jumpstarter.exporter.session import Session

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class Exporter(AbstractAsyncContextManager, Metadata):
    channel_factory: Callable[[], grpc.aio.Channel]
    device_factory: Callable[[], Driver]
    lease_name: str = field(init=False, default="")
    tls: TLSConfigV1Alpha1 = field(default_factory=TLSConfigV1Alpha1)
    grpc_options: dict[str, str] = field(default_factory=dict)

    async def __aexit__(self, exc_type, exc_value, traceback):
        controller = jumpstarter_pb2_grpc.ControllerServiceStub(await self.channel_factory())
        logger.info("Unregistering exporter with controller")
        await controller.Unregister(
            jumpstarter_pb2.UnregisterRequest(
                reason="TODO",
            )
        )

    async def __handle(self, path, endpoint, token, tls_config, grpc_options):
        try:
            async with await connect_unix(path) as stream:
                async with connect_router_stream(endpoint, token, stream, tls_config, grpc_options):
                    pass
        except Exception as e:
            logger.info("failed to handle connection: {}".format(e))

    @asynccontextmanager
    async def session(self):
        controller = jumpstarter_pb2_grpc.ControllerServiceStub(await self.channel_factory())
        with Session(
            uuid=self.uuid,
            labels=self.labels,
            root_device=self.device_factory(),
        ) as session:
            async with session.serve_unix_async() as path:
                async with grpc.aio.secure_channel(
                    f"unix://{path}", grpc.local_channel_credentials(grpc.LocalConnectionType.UDS)
                ) as channel:
                    response = await jumpstarter_pb2_grpc.ExporterServiceStub(channel).GetReport(empty_pb2.Empty())
                    logger.info("Registering exporter with controller")
                    await controller.Register(
                        jumpstarter_pb2.RegisterRequest(
                            labels=self.labels,
                            reports=response.reports,
                        )
                    )
                yield path

    async def handle(self, lease_name, tg):
        logger.info("Listening for incoming connection requests on lease %s", lease_name)

        listen_tx, listen_rx = create_memory_object_stream()

        async def listen(retries=5, backoff=3):
            retries_left = retries
            while True:
                try:
                    controller = jumpstarter_pb2_grpc.ControllerServiceStub(await self.channel_factory())
                    async for request in controller.Listen(jumpstarter_pb2.ListenRequest(lease_name=lease_name)):
                        await listen_tx.send(request)
                except Exception as e:
                    if retries_left > 0:
                        retries_left -= 1
                        logger.info(
                            "Listen stream interrupted, restarting in {}s, {} retries left: {}".format(
                                backoff, retries_left, e
                            )
                        )
                        await sleep(backoff)
                    else:
                        raise
                else:
                    retries_left = retries

        tg.start_soon(listen)

        async with self.session() as path:
            async for request in listen_rx:
                logger.info("Handling new connection request on lease %s", lease_name)
                tg.start_soon(
                    self.__handle, path, request.router_endpoint, request.router_token, self.tls, self.grpc_options
                )

    async def serve(self):  # noqa: C901
        # initial registration
        async with self.session():
            pass
        started = False
        status_tx, status_rx = create_memory_object_stream()

        async def status(retries=5, backoff=3):
            retries_left = retries
            while True:
                try:
                    controller = jumpstarter_pb2_grpc.ControllerServiceStub(await self.channel_factory())
                    async for status in controller.Status(jumpstarter_pb2.StatusRequest()):
                        await status_tx.send(status)
                except Exception as e:
                    if retries_left > 0:
                        retries_left -= 1
                        logger.info(
                            "Status stream interrupted, restarting in {}s, {} retries left: {}".format(
                                backoff, retries_left, e
                            )
                        )
                        await sleep(backoff)
                    else:
                        raise
                else:
                    retries_left = retries

        async with create_task_group() as tg:
            tg.start_soon(status)
            async for status in status_rx:
                if self.lease_name != "" and self.lease_name != status.lease_name:
                    self.lease_name = status.lease_name
                    logger.info("Lease status changed, killing existing connections")
                    tg.cancel_scope.cancel()
                    break
                self.lease_name = status.lease_name
                if not started and self.lease_name != "":
                    started = True
                    tg.start_soon(self.handle, self.lease_name, tg)
                if status.leased:
                    logger.info("Currently leased by %s under %s", status.client_name, status.lease_name)
                else:
                    logger.info("Currently not leased")



import shutil

from .utils import launch_shell


def test_launch_shell(tmp_path, monkeypatch):
    monkeypatch.setenv("SHELL", shutil.which("true"))
    exit_code = launch_shell(host=str(tmp_path / "test.sock"), context="remote", allow=["*"], unsafe=False)
    assert exit_code == 0

    monkeypatch.setenv("SHELL", shutil.which("false"))
    exit_code = launch_shell(host=str(tmp_path / "test.sock"), context="remote", allow=["*"], unsafe=False)
    assert exit_code == 1



# Reference: https://docs.djangoproject.com/en/5.0/_modules/django/utils/module_loading/#import_string

import sys
from fnmatch import fnmatchcase
from importlib import import_module


def cached_import(module_path, class_name):
    # Check whether module is loaded and fully initialized.
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = import_module(module_path)
    return getattr(module, class_name)


def import_class(class_path: str, allow: list[str], unsafe: bool):
    """
    Import a class by its full class path while checking
    the path matches the given allow list with unix style glob

    e.g. `import_class("example_package.some_module.fooclass", allow=["example_package.*"], unsafe=false)`
    is equivalent to `from example_package.some_module import FooClass; return FooClass`

    while `import_class("example_package.some_module.fooclass", allow=["notexample_package.*"], unsafe=false)`
    throws ImportError due to not matching the allow list
    """
    if not unsafe:
        if not any(fnmatchcase(class_path, pattern) for pattern in allow):
            raise ImportError(f"{class_path} doesn't match any of the allowed patterns")
    try:
        module_path, class_name = class_path.rsplit(".", 1)
    except ValueError as e:
        raise ImportError(f"{class_path} doesn't look like a class path") from e
    try:
        return cached_import(module_path, class_name)
    except AttributeError as e:
        raise ImportError(f"{module_path} doesn't have specified class {class_name}") from e



from contextlib import contextmanager
from functools import wraps


def blocking(f):
    @wraps(f)
    @contextmanager
    def wrapper(*args, **kwargs):
        with kwargs["client"].portal.wrap_async_context_manager(f(*args, **kwargs)) as res:
            yield res

    return wrapper
